'''
FIR filterbank, can be treated as a special case of LinearFilterbank, but an
optimisation is possible using buffered output by using FFT based convolution
as in HRTF.apply. To do this is slightly tricky because it needs to cache
previous inputs. For the moment, we implement it as a special case of
LinearFilterbank but later this will change to using the FFT method.
'''
from brian import *
from filterbank import *
from linearfilterbank import *

__all__ = ['FIRFilterbank', 'LinearFIRFilterbank', 'FFTFIRFilterbank']

class LinearFIRFilterbank(LinearFilterbank):
    def __init__(self, source, impulse_response, minimum_buffer_size=None):
        # if a 1D impulse response is given we apply it to every channel
        # Note that because we are using LinearFilterbank at the moment, this
        # means duplicating the impulse response. However, it could be stored
        # just once when we move to using FFT based convolution and in fact this
        # will save a lot of computation as the FFT only needs to be computed
        # once then.
        if len(impulse_response.shape)==1:
            impulse_response = repeat(reshape(impulse_response, (1, len(impulse_response))), source.nchannels, axis=0)
        # Automatically duplicate mono input to fit the desired output shape
        if impulse_response.shape[0]!=source.nchannels:
            if source.nchannels!=1:
                raise ValueError('Can only automatically duplicate source channels for mono sources, use RestructureFilterbank.')
            source = RestructureFilterbank(source, impulse_response.shape[0])
        # Implement it as a LinearFilterbank
        b = reshape(impulse_response, impulse_response.shape+(1,))
        a = zeros_like(b)
        a[:, 0, :] = 1
        LinearFilterbank.__init__(self, source, b, a)
        if minimum_buffer_size is not None:
            self.minimum_buffer_size = minimum_buffer_size

class FFTFIRFilterbank(Filterbank):
    def __init__(self, source, impulse_response, minimum_buffer_size=None):
        # if a 1D impulse response is given we apply it to every channel
        # Note that because we are using LinearFilterbank at the moment, this
        # means duplicating the impulse response. However, it could be stored
        # just once when we move to using FFT based convolution and in fact this
        # will save a lot of computation as the FFT only needs to be computed
        # once then.
        if len(impulse_response.shape)==1:
            impulse_response = repeat(reshape(impulse_response, (1, len(impulse_response))), source.nchannels, axis=0)
        # Automatically duplicate mono input to fit the desired output shape
        if impulse_response.shape[0]!=source.nchannels:
            if source.nchannels!=1:
                raise ValueError('Can only automatically duplicate source channels for mono sources, use RestructureFilterbank.')
            source = RestructureFilterbank(source, impulse_response.shape[0])
        Filterbank.__init__(self, source)

        self.input_cache = zeros((impulse_response.shape[1], self.nchannels))
        self.impulse_response = impulse_response
        self.fftcache_nmax = -1
        if minimum_buffer_size is None:
            minimum_buffer_size = 3*impulse_response.shape[1]
        self.minimum_buffer_size = minimum_buffer_size

    def buffer_init(self):
        Filterbank.buffer_init(self)
        self.input_cache[:] = 0

    # This version uses a single FFT/IFFT call, using the axis keyword, but it
    # doesn't appear to be any more efficient than looping, and uses much more
    # memory, although my tests weren't exhaustive.
#    def buffer_apply(self, input):
#        nmax = max(self.input_cache.shape[0]+input.shape[0], self.impulse_response.shape[1])
#        nmax = 2**int(ceil(log2(nmax)))
#        if self.fftcache_nmax!=nmax:
#            # impulse response: (nchannels, ir_length)
#            ir = zeros((self.nchannels, nmax))
#            ir[:, :self.impulse_response.shape[1]] = self.impulse_response
#            # fftcache: (ir_length, nchannels)
#            self.fftcache = fft(ir, n=nmax, axis=1).T
#            self.fftcache_nmax = nmax
#        fullinput = vstack((self.input_cache, input))
#        fullinput = vstack((fullinput, zeros((nmax-fullinput.shape[1], self.nchannels))))
#        # fullinput: (ir_length, nchannels)
#        fullinput_fft = fft(fullinput, n=nmax, axis=0)
#        fulloutput_fft = fullinput_fft*self.fftcache
#        fulloutput = ifft(fulloutput_fft, n=nmax, axis=0).real
#        output = fulloutput[self.input_cache.shape[0]:self.input_cache.shape[0]+input.shape[0]]
#        # update input cache
#        nic = self.input_cache.shape[0]
#        ni = input.shape[0]
#        #print ni, nic
#        if ni>=nic:
#            self.input_cache[:, :] = input[-nic:, :]
#        else:
#            self.input_cache[:-ni, :] = self.input_cache[ni:, :]
#            self.input_cache[-ni:, :] = input
#        return output
    
    def buffer_apply(self, input):
        output = zeros_like(input)
        nmax = max(self.input_cache.shape[0]+input.shape[0], self.impulse_response.shape[1])
        nmax = 2**int(ceil(log2(nmax)))
        if self.fftcache_nmax!=nmax:
            self.fftcache = []
        for i, (previnput, curinput, ir) in enumerate(zip(self.input_cache.T,
                                                          input.T,
                                                          self.impulse_response)):
            fullinput = hstack((previnput, curinput))
            # pad
            fullinput = hstack((fullinput, zeros(nmax-len(fullinput))))
            # apply fft
            if self.fftcache_nmax!=nmax:
                # recompute IR fft, first pad, then take fft, then store
                ir = hstack((ir, zeros(nmax-len(ir))))
                ir_fft = fft(ir, n=nmax)
                self.fftcache.append(ir_fft)
            else:
                ir_fft = self.fftcache[i]
            fullinput_fft = fft(fullinput, n=nmax)
            curoutput_fft = fullinput_fft*ir_fft
            curoutput = ifft(curoutput_fft)
            # unpad
            curoutput = curoutput[len(previnput):len(previnput)+len(curinput)]
            output[:, i] = curoutput.real
        if self.fftcache_nmax!=nmax:
            self.fftcache_nmax = nmax
        # update input cache
        nic = self.input_cache.shape[0]
        ni = input.shape[0]
        #print ni, nic
        if ni>=nic:
            self.input_cache[:, :] = input[-nic:, :]
        else:
            self.input_cache[:-ni, :] = self.input_cache[ni:, :]
            self.input_cache[-ni:, :] = input
        return output


class FIRFilterbank(Filterbank):
    '''
    Finite impulse response filterbank
    
    Initialisation parameters:
    
    ``source``
        Source sound or filterbank.
    ``impulse_response``
        Either a 1D array providing a single impulse response applied to every
        input channel, or a 2D array of shape ``(nchannels, ir_length)`` for
        ``ir_length`` the number of samples in the impulse response. Note that
        if you are using a multichannel sound ``x`` as a set of impulse responses,
        the array should be ``impulse_response=array(x.T)``.
    ``minimum_buffer_size=None``
        If specified, gives a minimum size to the buffer. By default, for the
        FFT convolution based implementation of ``FIRFilterbank``, the minimum
        buffer size will be ``3*ir_length``. For maximum efficiency with FFTs,
        ``buffer_size+ir_length`` should be a power of 2 (otherwise there will
        be some zero padding), and ``buffer_size`` should be as large as
        possible.
    '''
    def __init__(self, source, impulse_response, use_linearfilterbank=False,
                 minimum_buffer_size=None):
        if use_linearfilterbank:
            self.__class__ = LinearFIRFilterbank
        else:
            self.__class__ = FFTFIRFilterbank
        self.__init__(source, impulse_response,
                      minimum_buffer_size=minimum_buffer_size)



from numpy import *
import multiprocessing
import time
from multiprocessing import sharedctypes
from numpy import ctypeslib
import ctypes

def f(x, a, n):
    s = 0
    for _ in xrange(n):
        s += sum(a * x)
    return s

def test_with_copy_func((x, a, n)):
    return f(x, a, n)

def test_with_copy(N, M, complexity):
    x = ones(N)
    A = 1 + arange(M)
    pool = multiprocessing.Pool()
    args = [(x, a, complexity) for a in A]

    start = time.time()
    results = pool.map(test_with_copy_func, args)
    end = time.time()
    t_multiprocessing = end - start

    results = []
    start = time.time()
    for x, a, complexity in args:
        results.append(test_with_copy_func((x, a, complexity)))
    end = time.time()
    t_single = end - start

    print 'Test with copy'
    print '--------------'
    print 'Single:', t_single
    print 'Multiprocessing:', t_multiprocessing
    print 'S/M:', t_single / t_multiprocessing

def test_with_shared_func(x, a, complexity):
    x = ctypeslib.as_array(x)
    return f(x, a, complexity)

def test_with_shared(N, M, complexity):
    x = ones(N)
    x_sct = sharedctypes.Array(ctypes.c_double, x, lock=False)
    A = 1 + arange(M)

    #pool = multiprocessing.Pool()
    start = time.time()
    processes = [multiprocessing.Process(target=test_with_shared_func,
                                         args=(x_sct, a, complexity)) for a in A]
    for p in processes:
        p.start()
    for p in processes:
        p.join()
    end = time.time()
    t_multiprocessing = end - start

    results = []
    args = [(x, a, complexity) for a in A]
    start = time.time()
    for x, a, complexity in args:
        results.append(test_with_shared_func(x, a, complexity))
    end = time.time()
    t_single = end - start

    print 'Test with shared'
    print '----------------'
    print 'Single:', t_single
    print 'Multiprocessing:', t_multiprocessing
    print 'S/M:', t_single / t_multiprocessing

if __name__ == '__main__':
    N = 10000000
    M = 8
    complexity = 1
    print 'Array size:', N
    print 'Num processors:', M
    print 'Task complexity:', complexity
    print
    test_with_shared(N, M, complexity)
    print
    test_with_copy(N, M, complexity)



from brian import *

def testfromtutorial1c():
    '''Tests a behaviour from Tutorial 1c
    
    Solving the differential equation gives:
    
        V = El + (Vr-El) exp (-t/tau)
    
    Setting V=Vt at time t gives:
    
        t = tau log( (Vr-El) / (Vt-El) )
    
    If the simulator runs for time T, and fires a spike immediately
    at the beginning of the run it will then generate n spikes,
    where:
    
        n = [T/t] + 1
    
    If you have m neurons all doing the same thing, you get nm
    spikes. This calculation with the parameters above gives:
    
        t = 48.0 ms
        n = 21
        nm = 840
    
    As predicted.
    '''
    reinit_default_clock()
    tau = 20 * msecond        # membrane time constant
    Vt = -50 * mvolt          # spike threshold
    Vr = -60 * mvolt          # reset value
    El = -49 * mvolt          # resting potential (same as the reset)
    dV = 'dV/dt = -(V-El)/tau : volt # membrane potential'
    G = NeuronGroup(N=40, model=dV, threshold=Vt, reset=Vr)
    G.V = El
    M = SpikeMonitor(G)
    run(1 * second)
    assert M.nspikes == 840

if __name__ == '__main__':
    testfromtutorial1c()



'''
Spike propagation with pypar.

I am not sure how broadcasting works. Maybe it is not the better strategy.
Other option: send spikes in ring (n turns to transfer all spikes).

Documentation: http://datamining.anu.edu.au/~ole/pypar/DOC
'''

from brian import *
import pypar

# Identification
myid = pypar.rank() # id of this process
nproc = pypar.size() # number of processors
node = pypar.get_processor_name()

print "I am processor %d of %d on node %s" % (myid, nproc, node)

# Generation of spikes
myspikes = randint(100, size=randint(10))
print "I am sending", myspikes

# Broadcasting
spikes = [[] for _ in range(nproc)]
spikes[myid] = myspikes
nspikes = array([0])
for i in range(nproc):
    if i == myid: # that's me!
        # Create spikes
        nspikes[0] = len(myspikes)
    pypar.broadcast(nspikes, i)
    if i != myid: # This would be the virtual group
        spikes[i] = zeros(nspikes, dtype=int)
    pypar.broadcast(spikes[i], i)

print "I received", spikes
pypar.finalize()

#if myid == 0:
#    x=2*mV
#    pypar.send(x,1)
#    y=pypar.receive(1)
#    print "I got a big number:",y
#    v=zeros(10)
#    pypar.receive(1,buffer=v) # faster with arrays, does not copy the data
#    print "and now random numbers!"
#    print v
#    print "I am sending spikes!"
#    nspikes=array([3])
#    pypar.broadcast(nspikes,0)
#    spikes=randint(100,size=nspikes)
#    pypar.broadcast(spikes,0)
#else:
#    x=pypar.receive(0)
#    print "I received a message:",x
#    print "and I am doubling it"
#    pypar.send(2*x,0)
#    print "I send an array now"
#    pypar.send(randn(10),0,use_buffer=True)
#    nspikes=array([0])
#    pypar.broadcast(nspikes,0)
#    spikes=zeros(nspikes,dtype=int)
#    pypar.broadcast(spikes,0)
#    print "Broadcast result:"
#    print spikes
#pypar.finalize()



from brian import *
from filterbank import Filterbank,FunctionFilterbank,ControlFilterbank, CombinedFilterbank
from filterbanklibrary import *
from linearfilterbank import *
import warnings
from scipy.io import loadmat,savemat
from brian.hears import *


try:
    from scikits.samplerate import resample
    have_scikits_samplerate = True
except (ImportError, ValueError):
    have_scikits_samplerate = False
#print have_scikits_samplerate

def set_parameters(cf,param):
    
    parameters=dict()
    parameters['fc_LP_control']=800*Hz
    parameters['fc_LP_fb']=500*Hz
    parameters['fp1']=1.0854*cf-106.0034
    parameters['ta']=10**(log10(cf)*1.0230 + 0.1607)
    parameters['tb']=10**(log10(cf)*1.4292 - 1.1550) - 1000
    parameters['gain80']=10**(log10(cf)*0.5732 + 1.5220)
    parameters['rgain']=10**( log10(cf)*0.4 + 1.9)
    parameters['average_control']=0.3357
    parameters['zero_r']= array(-10**( log10(cf)*1.5-0.9 ))   
        
    if param: 
        if not isinstance(param, dict): 
            raise Error('given parameters must be a dict')
        for key in param.keys():
            if not parameters.has_key(key):
                raise Exception(key + ' is invalid key entry for given parameters')
            parameters[key] = param[key]
    parameters['nlgain']= (parameters['gain80'] - parameters['rgain'])/parameters['average_control']
    return parameters

##
def gain_groupdelay(binwidth,centerfreq,cf,tau):
    
    tmpcos = cos(2*pi*(centerfreq-cf)*binwidth)
    dtmp2 = tau*2.0/binwidth
    c1LP = (dtmp2-1)/(dtmp2+1)
    c2LP = 1.0/(dtmp2+1)
    tmp1 = 1+c1LP*c1LP-2*c1LP*tmpcos
    tmp2 = 2*c2LP*c2LP*(1+tmpcos)
    
    wb_gain = sqrt(tmp1/tmp2)
    
    grdelay = floor((0.5-(c1LP*c1LP-c1LP*tmpcos)/(1+c1LP*c1LP-2*c1LP*tmpcos))).astype(int)
    
    return wb_gain,grdelay

def get_taubm(cf,CAgain,taumax):
    bwfactor = 0.7;
    factor   = 2.5;
    ratio  = 10**(-CAgain/(20.0*factor))
    bmTaumax = taumax/bwfactor;
    bmTaumin = bmTaumax*ratio;   
    return bmTaumax,bmTaumin,ratio

def get_tauwb(cf,CAgain,order):
    ####
    ratio = 10**(-CAgain/(20.0*order))     #ratio of TauMin/TauMax according to the gain, order */
    ##Q10 = pow(10,0.4708*log10(cf/1e3)+0.5469); */ /* 75th percentile */
    Q10 = 10**(0.4708*log10(cf/1e3)+0.4664) #/* 50th percentile */
    ##Q10 = pow(10,0.4708*log10(cf/1e3)+0.3934); */ /* 25th percentile */
    bw     = cf/Q10;
    taumax = 2.0/(2*pi*bw);
    taumin   = taumax*ratio;
    return taumax,taumin


### function to initialize the chirp filters
class Chirp_Coefficients:
    
    def __init__(self,cf,taumax,samplerate,rsigma,fcohc):
        self.nch=len(cf)
        self.T = 1./ samplerate  
        self.TWOPI = 2*pi
        self.sigma0 = 1/taumax
        self.rsigma = rsigma
        self.fcohc = fcohc
        self.ipw    = 1.01*cf*self.TWOPI-50
        self.ipb    = 0.2343*self.TWOPI*cf-1104
        self.rpa    = pow(10, log10(cf)*0.9 + 0.55)+ 2000
        self.pzero  = pow(10,log10(cf)*0.7+1.6)+500
        
        self.order_of_pole = 10             
        self.half_order_pole = self.order_of_pole/2
        self.order_of_zero  = self.half_order_pole
        
        self.fs_bilinear = self.TWOPI*cf/tan(self.TWOPI*cf*self.T/2)
        self.fs_bilinear =tile(self.fs_bilinear.reshape(self.nch,-1),5)
        self.rzero       = -self.pzero
        self.CF = self.TWOPI*cf
        self.nch=len(self.CF)
        self.filt_a = zeros((len(cf),3,self.half_order_pole), order='F')
        self.filt_a[:,0,:] = 1
        self.filt_b = zeros((len(cf),3,self.half_order_pole), order='F')
        self.preal = zeros((self.nch,3))
        self.pimg = zeros((self.nch,6))
        
        self.preal = self.analog_poles_real(0*ones(self.nch),1*ones(self.nch))
        self.pimg = self.analog_poles_img()
        self.CFmat = tile(self.CF.reshape(self.nch,-1),5)
        self.rzeromat = tile(self.rzero.reshape(self.nch,-1),5)
        self.Cinitphase = sum(arctan(self.CFmat/(-self.rzeromat))\
          -arctan((self.CFmat-self.pimg[:,[0,2,1,0,1]])/(-self.preal[:,[0,2,1,0,1]]))\
                        -arctan((self.CFmat+self.pimg[:,[0,2,1,0,1]])/(-self.preal[:,[0,2,1,0,1]])),axis=1)
        self.CFmat10 = tile(self.CF.reshape(self.nch,-1),10)
        self.Cgain_norm = prod((self.CFmat10-self.pimg[:,[0,1,2,3,4,5,0,3,1,5]])**2+self.preal[:,[0,1,2,0,2,1,0,0,1,1]]**2,axis=1)
        self.norm_gain = sqrt(self.Cgain_norm)/sqrt(self.CF**2+self.rzero**2)**5
        self.norm_gain= tile(self.norm_gain.reshape(self.nch,-1),3)
        
    def return_coefficients(self):
        self.preal = self.analog_poles_real(self.rsigma,self.fcohc)
        self.phase = sum(-arctan((self.CFmat-self.pimg[:,[0,2,1,0,1]])/(-self.preal[:,[0,2,1,0,1]]))\
                            -arctan((self.CFmat+self.pimg[:,[0,2,1,0,1]])/(-self.preal[:,[0,2,1,0,1]])),axis=1)
        self.rzero = -self.CF/tan((self.Cinitphase-self.phase)/self.order_of_zero)
        self.rzero = tile(self.rzero.reshape(self.nch,-1),5)
        
        iord = [0,2,1,0,1]
        temp = (self.fs_bilinear-self.preal[:,iord])**2+self.pimg[:,iord]**2       
        self.filt_a[:,0,:] = 1
        self.filt_a[:,1,:] = -2*(self.fs_bilinear**2-self.preal[:,iord]**2-self.pimg[:,iord]**2)/temp            
        self.filt_a[:,2,:] = ((self.fs_bilinear+self.preal[:,iord])**2+self.pimg[:,iord]**2)/temp
        self.filt_b[:,0,:] = (-self.rzero+self.fs_bilinear)/temp
        self.filt_b[:,1,:] = (-2*self.rzero)/temp
        self.filt_b[:,2,:] = (-self.rzero-self.fs_bilinear)/temp
        self.filt_b[:,:,4] = self.norm_gain/4.*self.filt_b[:,:,4]                    
        return self.filt_b,self.filt_a

    def analog_poles_real(self,rsigma,fcohc):
        self.preal[:,0] = -self.sigma0*fcohc-rsigma  #0
        self.preal[:,1] = self.preal[:,0] - self.rpa #4
        self.preal[:,2] = (self.preal[:,0]+self.preal[:,1])*0.5 #2
        return self.preal
    
    def analog_poles_img(self):
        self.pimg[:,0] = self.ipw #0
        self.pimg[:,1] = self.pimg[:,0] - self.ipb #4
        self.pimg[:,2] = (self.pimg[:,0]+self.pimg[:,1])*0.5 #2
        self.pimg[:,3] = -self.pimg[:,0]  #1
        self.pimg[:,4] = -self.pimg[:,2]  #3
        self.pimg[:,5] = -self.pimg[:,1]  #5
        return self.pimg
    
#### controlers #####         
#definition of the class updater for the signal path bandpass filter
class Filter_Update: 
    def __init__(self, target,c1_coefficients,samplerate,cf,bmTaumax,bmTaumin,cohc,TauWBMax,TauWBMin):
        self.bmTaumax = bmTaumax
        self.bmTaumin = bmTaumin
        self.target=target
        self.samplerate=samplerate
        self.cf=atleast_1d(cf)
        self.cohc = cohc
        self.TauWBMax = TauWBMax
        self.TauWBMin = TauWBMin
        bmplace = 11.9 * log10(0.80 + cf / 456.0)
        self.centerfreq = 456.0*(pow(10,(bmplace+1.2)/11.9)-0.80)
        self.c1_coefficients=c1_coefficients
        self.rsigma=[]
    def __call__(self,input):  
      

        tmptauc1 = input[-1,:]
        tauc1    = self.cohc*(tmptauc1-self.bmTaumin)+self.bmTaumin
        #signal path update
        self.c1_coefficients.rsigma   = 1/tauc1-1/self.bmTaumax
#        self.target[0].filt_b,self.target[0].filt_a = self.c1_coefficients.return_coefficients() 
#        #control path update
#        tauwb = self.TauWBMax+(tauc1-self.bmTaumax)*(self.TauWBMax-self.TauWBMin)/(self.bmTaumax-self.bmTaumin)
#        [wb_gain,self.grdelay] = gain_groupdelay(1./self.samplerate,self.centerfreq,self.cf,tauwb);
#        
#        grd[n] = grdelay 
#        if ((grd[n]+n)<totalstim)
#             tmpgain[grd[n]+n] = wb_gain
#        if (tmpgain[n] == 0)
#            tmpgain[n] = lasttmpgain
#        wbgain      = tmpgain[n]
#        lasttmpgain = wbgain
        #[self.target[1].filt_b,self.target[1].filt_a] = ...
        
def IHC_transduction(x,slope,asym,sign): 
    corner    = 80
    strength  = 20.0e6/10**(corner/20)     
    xx = sign*log(1.0+strength*abs(x))*slope
    ind = x<0
    splx   = 20*log10(-x[ind]/20e-6);
    asym_t = asym-(asym-1)/(1+exp(splx/5.0));
    xx[ind] = -1/asym_t*xx[ind]
    return xx


def boltzman(x,asym,s0,s1,x1):
    shift = 1.0/(1.0+asym) # asym is the ratio of positive Max to negative Max*/
    x0    = s0*log((1.0/shift-1)/(1+exp(x1/s1)))
    out1 = 1.0/(1.0+exp(-(x-x0)/s0)*(1.0+exp(-(x-x1)/s1)))-shift
    return out1/(1-shift)


def OHC_transduction(x,taumin,taumax,asym):
    minR = 0.05*ones(len(taumin))
    R  = taumin/taumax
    ind = R<minR
    minR[ind] =  0.5*R[ind]  
    dc = (asym-1)/(asym+1.0)/2.0-minR
    R1 = R-minR
    s0 = -dc/log(R1/(1-minR))
    #This is for new nonlinearity
    x1  = abs(x)
    out = zeros_like(x)
    for ix in xrange(len(x)):
        out[ix,:] = taumax*(minR+(1.0-minR)*exp(-x1[ix,:]/s0))
        ind = out[ix,:]<taumin
        out[ix,ind] = taumin[ind]
        ind = out[ix,:]>taumax
        out[ix,ind] = taumax[ind]
    return out 


class LowPass_filter(LinearFilterbank):
    def __init__(self,source,cf,fc,gain,order):
        nch = len(cf)
        TWOPI = 2*pi
        self.samplerate =  source.samplerate
        c = 2.0 * self.samplerate
        c1LP = ( c/Hz - TWOPI*fc ) / ( c/Hz + TWOPI*fc )
        c2LP = TWOPI*fc/Hz / (TWOPI*fc + c/Hz)
        
        b_temp = array([c2LP,c2LP])
        a_temp = array([1,-c1LP])
        
        filt_b = tile(b_temp.reshape([2,1]),[nch,1,order])               
        filt_a = tile(a_temp.reshape([2,1]),[nch,1,order]) 
        filt_b[:,:,0] = filt_b[:,:,0]*gain

        LinearFilterbank.__init__(self, source, filt_b, filt_a)

class ZILANY(CombinedFilterbank):
    '''
    Class implementing the nonlinear auditory filterbank model as described in
    Tan, G. and Carney, L., 
    "A phenomenological model for the responses of auditory-nerve
    fibers. II. Nonlinear tuning with a frequency glide", JASA 2003.
    
    The model consists of a control path and a signal path. The control path
    controls both its own bandwidth via a feedback
    loop and also the bandwidth of the signal path. 
    
    Initialised with arguments:
    
    ``source``
        Source of the cochlear model.
        
    ``cf``
        List or array of center frequencies.
        
    ``update_interval``
        Interval in samples controlling how often the band pass filter of the
        signal pathway is updated. Smaller values are more accurate but
        increase the computation time.
        
    ``param``
        Dictionary used to overwrite the default parameters given in the
        original paper. 
    '''
    
    def __init__(self, source,cf,update_interval,param={}):
        file="/home/bertrand/Data/MatlabProg/brian_hears/ZilanyCarney-JASAcode-2009/wbout.mat"
        X=loadmat(file,struct_as_record=False)
        wbout = Sound(X['wbout'].flatten())
        wbout.samplerate = 100*kHz

        CombinedFilterbank.__init__(self, source)
#        source = self.get_modified_source()
        
        cf = atleast_1d(cf)
        nbr_cf=len(cf)
        samplerate=source.samplerate

        parameters=set_parameters(cf,param)
#        if int(source.samplerate)!=50000:
#            warnings.warn('To use the PMFR cochlear model the sample rate should be 50kHz')
#            if not have_scikits_samplerate:
#                raise ImportError('To use the PMFR cochlear model the sample rate should be 50kHz and scikits.samplerate package is needed for resampling')               
#            #source=source.resample(50*kHz)
#            warnings.warn('The input to the PMFR cochlear model has been resampled to 50kHz')

        cohc =1 # ohc scaling factor: 1 is normal OHC function; 0 is complete OHC dysfunction
        cihc = 1 # i ihc scaling factor: 1 is normal IHC function; 0 is complete IHC dysfunction
        bmplace = 11.9 * log10(0.80 + cf / 456.0)
        centerfreq = 456.0*(pow(10,(bmplace+1.2)/11.9)-0.80)
        CAgain = minimum(60,maximum(15*ones(len(cf)),52/2*(tanh(2.2*log10(cf/600)+0.15)+1)))
        
        
        # Parameters for the control-path wideband filter =======*/
        bmorder = 3;
        Taumax,Taumin = get_tauwb(cf,CAgain,bmorder)
        taubm   = cohc*(Taumax-Taumin)+Taumin;
        ratiowb = Taumin/Taumax;
        
        #====== Parameters for the signal-path C1 filter ======*/
        bmTaumax,bmTaumin,ratiobm = get_taubm(cf,CAgain,Taumax)    
        bmTaubm  = cohc*(bmTaumax-bmTaumin)+bmTaumin;
        fcohc    = bmTaumax/bmTaubm;
        
        # Parameters for the control-path wideband filter =======*/
        wborder = 3
        TauWBMax = Taumin+0.2*(Taumax-Taumin);
        TauWBMin = TauWBMax/Taumax*Taumin;
        tauwb    = TauWBMax+(bmTaubm-bmTaumax)*(TauWBMax-TauWBMin)/(bmTaumax-bmTaumin);
        [wbgain,grdelay] = gain_groupdelay(1./self.samplerate,centerfreq,cf,tauwb)
        
        # Nonlinear asymmetry of OHC function and IHC C1 transduction function*/
        ohcasym  = 7.0    
        ihcasym  = 3.0
        
        
#        ##### Control Path ####
#        print tauwb*pi*centerfreq
#        gt_control = BiQuadratic(source, centerfreq,tauwb*2*pi*centerfreq)
        gt_control = ApproximateGammatone(source, centerfreq, 1./(2*pi*tauwb), order=3) 
#        gt_control = LinearGammachirp(source, centerfreq,tauwb, c=0)
#        gt_control = Gammatone(source, centerfreq, b=1./(pi*tauwb))
#        def wb_gain(x,cf=array([1000]),tauwb=array([.0003]),TauWBMax=array([.0003]),wborder=3):
#            print x.shape,cf.shape
#            out=zeros_like(x)
#            for ix in xrange(len(x)):
#                out[ix,:] = (tauwb/TauWBMax)**wborder*x[ix,:]*10e3*maximum(ones(len(cf)),cf/5e3)
#            return out
#        gt_control1 = FunctionFilterbank(gt_control,wb_gain,cf=cf,tauwb=tauwb,TauWBMax=TauWBMax,wborder=wborder) 
        max_temp = (tauwb/TauWBMax)**wborder*10e3*maximum(ones(len(cf)),cf/5e3)
        gt_control1 = FunctionFilterbank(gt_control,lambda x:x*max_temp)

        # first non linearity of control path
        NL1_control=FunctionFilterbank(gt_control1,boltzman,asym=ohcasym,s0=12.0,s1=5.0,x1=5.0)        
#        #control low pass filter (its output will be used to control the signal path)
        LP_control = LowPass_filter(NL1_control,cf,600,1.0,2)        
#        # second non linearity of control path
        NL2_control=FunctionFilterbank(LP_control,OHC_transduction,taumin=bmTaumin, taumax=bmTaumax, asym=ohcasym) 
#           
#        #### C1  ####
        rsigma = zeros(len(cf))
        c1_coefficients = Chirp_Coefficients(cf, bmTaumax, samplerate, rsigma,1*ones(len(cf)))
        [filt_b,filt_a] = c1_coefficients.return_coefficients()
        C1_filter = LinearFilterbank(source,filt_b,filt_a)
        C1_IHC = FunctionFilterbank(C1_filter,IHC_transduction,slope = 0.1,asym = ihcasym,sign=1) 
        #### C2  ####
        c2_coefficients = Chirp_Coefficients(cf, bmTaumax, samplerate, 0*ones(len(cf)),1./ratiobm)
        [filt_b,filt_a] = c2_coefficients.return_coefficients()
        C2_filter = LinearFilterbank(source,filt_b,filt_a)
        gain_temp = cf**2/2e4
        C2_IHC_pre = FunctionFilterbank(C2_filter,lambda x:x*abs(x)*gain_temp)
        C2_IHC = FunctionFilterbank(C2_IHC_pre,IHC_transduction,slope = 0.2,asym = 1.0,sign=-1) 
#        
        C_IHC = C1_IHC + C2_IHC
#        
        C_IHC_lp = LowPass_filter(C_IHC,cf,3000,1.0,7)
 
        #controlers definition
        updater=Filter_Update([C1_filter],c1_coefficients,samplerate,cf,bmTaumax,bmTaumin,cohc,TauWBMax,TauWBMin) #instantiation of the updater for the control path
        output = ControlFilterbank(C1_filter, NL2_control, [C1_filter],updater, update_interval)  #controler for the band pass filter of the control path
        self.set_output(output)

        
        #line 354


"""
An example of the :class:`TimedArray` class used for applying input currents
to neurons.
"""
from brian import *

N = 5
duration = 100 * ms
Vr = -60 * mV
Vt = -50 * mV
tau = 10 * ms
Rmin = 1 * Mohm
Rmax = 10 * Mohm
freq = 50 * Hz
k = 10 * nA

eqs = '''
dV/dt = (-(V-Vr)+R*I)/tau : volt
R : ohm
I : amp
'''

G = NeuronGroup(N, eqs, reset='V=Vr', threshold='V>Vt')
G.R = linspace(Rmin, Rmax, N)

t = linspace(0 * second, duration, int(duration / defaultclock.dt))
I = clip(k * sin(2 * pi * freq * t), 0, Inf)
G.I = TimedArray(I)

M = MultiStateMonitor(G, record=True)

run(duration)

subplot(211)
M['I'].plot()
ylabel('I (amp)')
subplot(212)
M['V'].plot()
ylabel('V (volt)')
show()



# -*- coding:utf-8 -*-
#class UserComputedConnectionMatrix(ConnectionMatrix):
#    '''
#    A computed connection matrix defined by a user-specified function
#    
#    Normally this matrix will be initialised by passing the class
#    object to the :class:`Connection` object. In the initialisation
#    of the :class:`Connection` specify ``structure=UserComputedConnectionMatrix``
#    and add the keyword ``row_func=...``, e.g.::
#    
#        def f(i):
#            return max_weight*ones(N)/(1+(arange(N)-i)**2)
#        C = Connection(G1, G2, structure=UserComputedConnectionMatrix, row_func=f)
#    
#    Initialisation arguments:
#    
#    ``dims``
#        The pair ``(N,M)`` specifying the dimensions of the matrix.
#    ``row_func``
#        The function ``f(i)`` which returns an array of length ``M``,
#        the weight matrix for row ``i``. Note that you are responsible
#        for making sure the function returns consistent results (so
#        random functions should be initialised with a seed based on
#        the row ``i``).
#    
#    **Limitations**
#    
#    This type of connection matrix cannot be changed during a run, and
#    cannot be used with methods like :class:`Connection.connect_random`.
#    
#    **Efficiency considerations**
#    
#    This connection matrix is for dense connectivity, if the connectivity
#    is sparse you might get better performance with :class:`UserComputedSparseConnectionMatrix`.
#    '''
#    def __init__(self, dims, row_func):
#        self.sourcelen, self.targetlen = dims
#        self.row_func = row_func
#        
#    def get_row(self, i):
#        return self.row_func(i)
#    
#    def __getitem__(self, item):
#        if isinstance(item,int):
#            return self.get_row(item)
#        if isinstance(item,tuple):
#            if len(item)==2:
#                item_i, item_j = item
#                if isinstance(item_i, int) and isinstance(item_j, slice):
#                    if is_colon_slice(item_j):
#                        return self.get_row(item_i)
#        raise ValueError('Only "i,:" indexing supported.')
#
#def random_sparse_row_func(N, p, weight=1., initseed=None):
#    '''
#    Returns a random connectivity ``row_func`` for use with :class:`UserComputedSparseConnectionMatrix`
#    
#    Gives equivalent output to the :meth:`Connection.connect_random` method.
#    
#    Arguments:
#    
#    ``N``
#        The number of target neurons.
#    ``p``
#        The probability of a synapse.
#    ``weight``
#        The connection weight (must be a single value).
#    ``initseed``
#        The initial seed value (for reproducible results).
#    '''
#    if initseed is None:
#        initseed = pyrandom.randint(100000,1000000) # replace this
#    myrange = numpy.arange(N, dtype=int)
#    def row_func(i):
#        pyrandom.seed(initseed+int(i))
#        scirandom.seed(initseed+int(i))
#        k = scirandom.binomial(N, p, 1)[0]
#        return (pyrandom.sample(myrange,k), weight)
#    return row_func
#
#class UserComputedSparseConnectionMatrix(ConnectionMatrix):
#    '''
#    A computed sparse connection matrix defined by a user-specified function
#    
#    Normally this matrix will be initialised by passing the class
#    object to the :class:`Connection` object. In the initialisation
#    of the :class:`Connection` specify ``structure=UserComputedSparseConnectionMatrix``
#    and add the keyword ``row_func=...``, e.g.::
#    
#        def f(i):
#            if 0<i<N-1:
#                return ([i-1,i+1], weight*ones(2))
#            elif i>0:
#                return ([i-1], weight*ones(1))
#            else:
#                return ([i+1], weight*ones(1))
#        C = Connection(G1, G2, structure=UserComputedSparseConnectionMatrix, row_func=f)
#    
#    Initialisation arguments:
#    
#    ``dims``
#        The pair ``(N,M)`` specifying the dimensions of the matrix.
#    ``row_func``
#        The function ``f(i)`` which for a row ``i`` returns a pair ``(indices, values))``
#        consisting of a list or array ``indices`` with the indices of the
#        nonzero elements of the row, and an array of the same length ``values``
#        giving the weight matrix for those indices. Note that you are responsible
#        for making sure the function returns consistent results (so
#        random functions should be initialised with a seed based on
#        the row ``i``).
#    
#    **Limitations**
#    
#    This type of connection matrix cannot be changed during a run, and
#    cannot be used with methods like :class:`Connection.connect_random`.
#    
#    **Efficiency considerations**
#    
#    This connection matrix is for sparse connectivity, if the connectivity
#    is dense you might get better performance with :class:`UserComputedConnectionMatrix`.
#    '''
#    def __init__(self, dims, row_func):
#        self.sourcelen, self.targetlen = dims
#        self.row_func = row_func
#        self.cur_row = numpy.zeros(dims[1])
#        
#    def add_row(self,i,X):
#        indices, values = self.row_func(i)
#        X[indices]+=values
#        
#    def add_scaled_row(self,i,X,factor):
#        # modulation may not work? need factor[self.rows[i]] here? is factor a number or an array?
#        X[indices]+=factor*values
#        
#    def get_row(self, i):
#        indices, values = self.row_func(i)
#        self.cur_row[:] = 0.0
#        self.cur_row[indices] = values
#        return self.cur_row
#    
#    def __getitem__(self, item):
#        if isinstance(item,int):
#            return self.get_row(item)
#        if isinstance(item,tuple):
#            if len(item)==2:
#                item_i, item_j = item
#                if isinstance(item_i, int) and isinstance(item_j, slice):
#                    if is_colon_slice(item_j):
#                        return self.get_row(item_i)
#        raise ValueError('Only "i,:" indexing supported.')



'''
There's a bug in CSTDP so that if you have more target neurons than sources,
it will segfault, suggesting that one of the values passed to the C code is
wrong.
'''
from brian import *

set_global_preferences(useweave=True, usecstdp=True)
log_level_debug()

Nin = 100
Nout = 200
duration = 40 * second

Fmax = 50 * Hz

tau = 10 * ms
taue = 2 * ms
taui = 5 * ms
sigma = 0.#0.4
# Note that factor (tau/taue) makes integral(v(t)) the same when the connection
# acts on ge as if it acted directly on v.
eqs = Equations('''
#dv/dt = -v/tau + sigma*xi/(2*tau)**.5 : 1
dv/dt = (-v+(tau/taue)*ge-(tau/taui)*gi)/tau + sigma*xi/(2*tau)**.5 : 1
dge/dt = -ge/taue : 1
dgi/dt = -gi/taui : 1
excitatory = ge
inhibitory = gi
''')
reset = 0
threshold = 1
refractory = 0 * ms
taup = 5 * ms
taud = 5 * ms
Ap = .1
Ad = -Ap * taup / taud * 1.2
wmax_ff = 0.1
wmax_rec = wmax_ff
wmax_inh = wmax_rec

width = 0.2

recwidth = 0.2

Gin = PoissonGroup(Nin)
Gout = NeuronGroup(Nout, eqs, reset=reset, threshold=threshold, refractory=refractory)
ff = Connection(Gin, Gout, 'excitatory', structure='dense')
for i in xrange(Nin):
    ff[i, :] = (rand(Nout) > .5) * wmax_ff
rec = Connection(Gout, Gout, 'excitatory')
for i in xrange(Nout):
    d = abs(float(i) / Nout - linspace(0, 1, Nout))
    d[d > .5] = 1. - d[d > .5]
    dsquared = d ** 2
    prob = exp(-dsquared / (2 * recwidth ** 2))
    prob[i] = -1
    inds = (rand(Nout) < prob).nonzero()[0]
    w = rand(len(inds)) * wmax_rec
    rec[i, inds] = w

inh = Connection(Gout, Gout, 'inhibitory', sparseness=1, weight=wmax_inh)

stdp_ff = ExponentialSTDP(ff, taup, taud, Ap, Ad, wmax=wmax_ff)
#stdp_rec = ExponentialSTDP(rec, taup, taud, Ap, Ad, wmax=wmax_rec)

run(0 * ms)
@network_operation(clock=EventClock(dt=20 * ms))
def stimulation():
    Gin.rate = Fmax * exp(-(linspace(0, 1, Nin) - rand())**2 / (2 * width ** 2))

run(1*second, report='stderr')



from multimap import *
import time

def fun(x):
    return x ** 2

# max_parallel is defined in multimap.py and contains the maximum number of jobs that can
# run simultaneously.
args = range(max_parallel)

# Launch the jobs on the cloud and returns the job ids used to retrieve the results later.
jids = multimap(fun, args)

# Prints the status of the jobs after 10 seconds
time.sleep(10)
print status(jids)

# Retrieves the results (BLOCKING call)
results = retrieve(jids)
print results




"""
CUBA optimising
"""

import BrianNoUnits
from brian import *
from brian.stdunits import *
from brian.globalprefs import *
from scipy import rand
import pylab
import time
import c_profile
import profile

def main():
    taum = 20 * ms
    taue = 5 * ms
    taui = 10 * ms
    Vt = -50 * mV
    Vr = -60 * mV
    dv = lambda v, ge, gi: (ge + gi - (v + 49 * mV)) / taum
    dge = lambda v, ge, gi:-ge / taue
    dgi = lambda v, ge, gi:-gi / taui
    P = NeuronGroup(16000, model=(dv, dge, dgi), threshold=Vt, \
                  reset=Vr, init=(0 * volt, 0 * volt, 0 * volt))
    Pe = P.subgroup(12800)
    Pi = P.subgroup(3200)
    Ce = Connection(Pe, P, 1)
    Ci = Connection(Pi, P, 2)
    we = (60 * 0.27 / 10) * mV # excitatory synaptic weight (voltage)
    wi = (-20 * 4.5 / 10) * mV # inhibitory synaptic weight
    Ce.connect_random(Pe, P, 0.02, weight=we)
    Ci.connect_random(Pi, P, 0.02, weight=wi)
    P['v'] = Vr + rand(1, len(P)) * (Vt - Vr)    # Or P.S[0]=...

    net = MagicNetwork()
    start = time.time()
    net.run(10 * second)
    print time.time() - start


set_global_preferences(useweave=True)

main()
#c_profile.run('main()')
#main()



'''
Interface example
Install cherrypy for this example
Then run the script and go to http://localhost:8080 on your web browser
You can use cherrypy to write html interfaces to your code.
'''

from brian import *
import cherrypy
import os.path

# The server is defined here
class MyInterface(object):
    @cherrypy.expose
    def index(self): # redirect to the html page we wrote
        return '<meta HTTP-EQUIV="Refresh" content="0;URL=index.html">'

    @cherrypy.expose
    def runscript(self, we="1.62", wi="-9", **kwd): # 'runscript' is the script name
        # we and wi are the names of form fields
        we = float(we)
        wi = float(wi)
        # From minimalexample
        reinit_default_clock()
        eqs = '''
        dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
        dge/dt = -ge/(5*ms) : volt
        dgi/dt = -gi/(10*ms) : volt
        '''
        P = NeuronGroup(4000, model=eqs, threshold= -50 * mV, reset= -60 * mV)
        P.v = -60 * mV + 10 * mV * rand(len(P))
        Pe = P.subgroup(3200)
        Pi = P.subgroup(800)
        Ce = Connection(Pe, P, 'ge')
        Ci = Connection(Pi, P, 'gi')
        Ce.connect_random(Pe, P, 0.02, weight=we * mV)
        Ci.connect_random(Pi, P, 0.02, weight=wi * mV)
        M = SpikeMonitor(P)
        run(.5 * second)
        clf()
        raster_plot(M)
        savefig('image.png')
        # Redirect to the html page we wrote
        return '<meta HTTP-EQUIV="Refresh" content="0;URL=results.html">'

# Set the directory for static files
current_dir = os.path.dirname(os.path.abspath(__file__))
conf = {'/': {'tools.staticdir.on':True,
              'tools.staticdir.dir':current_dir}}

# Start the server
cherrypy.quickstart(MyInterface(), config=conf)



# Matrix product testing
from scipy import *
from scipy import weave
from time import time
from scipy import linalg
import numpy

gemm, = linalg.blas.get_blas_funcs(['gemm'])

iterations = 10000
N = 4000
m = 3

A = array(rand(m, m))
M = array(rand(m, N))
B = array(zeros((m, N)))
V = array(rand(m, 1))

t1 = time()
for _ in xrange(iterations):
    B = numpy.core._dotblas.dot(A, M)
    #B=gemm(1.0,A,M,0,M,0,0,overwrite_c=0)
t2 = time()
print t2 - t1



import time
from numpy import *
from numpy.random import *
from fastexp import *

x = array([1, 2, 3], dtype=float)
y = exp(x)
z = zeros(len(x))
fastexp(x, z)
print y
print z

x = randn(10000000)
y = empty(x.shape)

t = time.time()
z = exp(x)
print "exp", time.time() - t

t = time.time()
fastexp(x, y)
print "fastexp", time.time() - t



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 1.3.36
#
# Don't modify this file, modify the SWIG interface instead.
# This file is compatible with both classic and new-style classes.

import _ccircular
import new
new_instancemethod = new.instancemethod
try:
    _swig_property = property
except NameError:
    pass # Python < 2.2 doesn't have 'property'.
def _swig_setattr_nondynamic(self,class_type,name,value,static=1):
    if (name == "thisown"): return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'PySwigObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name,None)
    if method: return method(self,value)
    if (not static) or hasattr(self,name):
        self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)

def _swig_setattr(self,class_type,name,value):
    return _swig_setattr_nondynamic(self,class_type,name,value,0)

def _swig_getattr(self,class_type,name):
    if (name == "thisown"): return self.this.own()
    method = class_type.__swig_getmethods__.get(name,None)
    if method: return method(self)
    raise AttributeError,name

def _swig_repr(self):
    try: strthis = "proxy of " + self.this.__repr__()
    except: strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

import types
try:
    _object = types.ObjectType
    _newclass = 1
except AttributeError:
    class _object : pass
    _newclass = 0
del types


class CircularVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, CircularVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, CircularVector, name)
    __swig_setmethods__["X"] = _ccircular.CircularVector_X_set
    __swig_getmethods__["X"] = _ccircular.CircularVector_X_get
    if _newclass:X = _swig_property(_ccircular.CircularVector_X_get, _ccircular.CircularVector_X_set)
    __swig_setmethods__["cursor"] = _ccircular.CircularVector_cursor_set
    __swig_getmethods__["cursor"] = _ccircular.CircularVector_cursor_get
    if _newclass:cursor = _swig_property(_ccircular.CircularVector_cursor_get, _ccircular.CircularVector_cursor_set)
    __swig_setmethods__["n"] = _ccircular.CircularVector_n_set
    __swig_getmethods__["n"] = _ccircular.CircularVector_n_get
    if _newclass:n = _swig_property(_ccircular.CircularVector_n_get, _ccircular.CircularVector_n_set)
    __swig_setmethods__["retarray"] = _ccircular.CircularVector_retarray_set
    __swig_getmethods__["retarray"] = _ccircular.CircularVector_retarray_get
    if _newclass:retarray = _swig_property(_ccircular.CircularVector_retarray_get, _ccircular.CircularVector_retarray_set)
    def __init__(self, *args): 
        this = _ccircular.new_CircularVector(*args)
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _ccircular.delete_CircularVector
    __del__ = lambda self : None;
    def reinit(*args): return _ccircular.CircularVector_reinit(*args)
    def advance(*args): return _ccircular.CircularVector_advance(*args)
    def __len__(*args): return _ccircular.CircularVector___len__(*args)
    def __getitem__(*args): return _ccircular.CircularVector___getitem__(*args)
    def __setitem__(*args): return _ccircular.CircularVector___setitem__(*args)
    def __getslice__(*args): return _ccircular.CircularVector___getslice__(*args)
    def get_conditional(*args): return _ccircular.CircularVector_get_conditional(*args)
    def __setslice__(*args): return _ccircular.CircularVector___setslice__(*args)
    def __repr__(*args): return _ccircular.CircularVector___repr__(*args)
    def __str__(*args): return _ccircular.CircularVector___str__(*args)
    def expand(*args): return _ccircular.CircularVector_expand(*args)
CircularVector_swigregister = _ccircular.CircularVector_swigregister
CircularVector_swigregister(CircularVector)

class SpikeContainer(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, SpikeContainer, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, SpikeContainer, name)
    __swig_setmethods__["S"] = _ccircular.SpikeContainer_S_set
    __swig_getmethods__["S"] = _ccircular.SpikeContainer_S_get
    if _newclass:S = _swig_property(_ccircular.SpikeContainer_S_get, _ccircular.SpikeContainer_S_set)
    __swig_setmethods__["ind"] = _ccircular.SpikeContainer_ind_set
    __swig_getmethods__["ind"] = _ccircular.SpikeContainer_ind_get
    if _newclass:ind = _swig_property(_ccircular.SpikeContainer_ind_get, _ccircular.SpikeContainer_ind_set)
    __swig_setmethods__["remaining_space"] = _ccircular.SpikeContainer_remaining_space_set
    __swig_getmethods__["remaining_space"] = _ccircular.SpikeContainer_remaining_space_get
    if _newclass:remaining_space = _swig_property(_ccircular.SpikeContainer_remaining_space_get, _ccircular.SpikeContainer_remaining_space_set)
    def __init__(self, *args): 
        this = _ccircular.new_SpikeContainer(*args)
        try: self.this.append(this)
        except: self.this = this
    __swig_destroy__ = _ccircular.delete_SpikeContainer
    __del__ = lambda self : None;
    def reinit(*args): return _ccircular.SpikeContainer_reinit(*args)
    def push(*args): return _ccircular.SpikeContainer_push(*args)
    def lastspikes(*args): return _ccircular.SpikeContainer_lastspikes(*args)
    def __getitem__(*args): return _ccircular.SpikeContainer___getitem__(*args)
    def get_spikes(*args): return _ccircular.SpikeContainer_get_spikes(*args)
    def __getslice__(*args): return _ccircular.SpikeContainer___getslice__(*args)
    def __repr__(*args): return _ccircular.SpikeContainer___repr__(*args)
    def __str__(*args): return _ccircular.SpikeContainer___str__(*args)
SpikeContainer_swigregister = _ccircular.SpikeContainer_swigregister
SpikeContainer_swigregister(SpikeContainer)






# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
"""
Global preferences for Brian
----------------------------

The following global preferences have been defined:

"""
# the global preferences referred to above are automatically
# added to the docstring when they are defined via
# define_global_preference
__docformat__ = "restructuredtext en"

__all__ = ['set_global_preferences', 'get_global_preference', 'exists_global_preference', 'define_global_preference']

import sys
from utils.documentation import *

globalprefdocs = ""


class BrianGlobalPreferences:
    pass
g_prefs = BrianGlobalPreferences()


def set_global_preferences(**kwds):
    """Set global preferences for Brian
    
    Usage::
    
        ``set_global_preferences(...)``
    
    where ... is a list of keyword assignments.
    """
    for k, v in kwds.iteritems():
        g_prefs.__dict__[k] = v


def get_global_preference(k):
    """Get the value of the named global preference
    """
    return g_prefs.__dict__[k]

def exists_global_preference(k):
    """Determine if named global preference exists
    """
    return hasattr(g_prefs, k)

def define_global_preference(k, defaultvaluedesc, desc):
    """Define documentation for a new global preference
    
    Arguments:
    
    ``k``
        The name of the preference (a string)
    ``defaultvaluedesc``
        A string description of the default value
    ``desc``
        A multiline description of the preference in
        docstring format
    """
    global globalprefdocs
    globalprefdocs += '``' + k + ' = ' + defaultvaluedesc + '``\n'
    globalprefdocs += flattened_docstring(desc, numtabs=1)



'''
Tutorial 1g: Recording membrane potentials
******************************************

In the previous part of this tutorial, we plotted a raster plot of
the firing times of the network. In this tutorial, we introduce
a way to record the value of the membrane potential for a neuron
during the simulation, and plot it. We continue as before:
'''
from brian import *

tau = 20 * msecond        # membrane time constant
Vt = -50 * mvolt          # spike threshold
Vr = -60 * mvolt          # reset value
El = -49 * mvolt          # resting potential (same as the reset)
psp = 0.5 * mvolt         # postsynaptic potential size

G = NeuronGroup(N=40, model='dV/dt = -(V-El)/tau : volt',
              threshold=Vt, reset=Vr)

C = Connection(G, G)
C.connect_random(sparseness=0.1, weight=psp)
'''
This time we won't record the spikes.

Recording states
~~~~~~~~~~~~~~~~

Now we introduce a second type of monitor, the :class:`StateMonitor`.
The first argument is the group to monitor, and the second is
the state variable to monitor. The keyword ``record`` can be
an integer, list or the value ``True``. If it is an integer ``i``,
the monitor will record the state of the variable for neuron ``i``.
If it's a list of integers, it will record the states for
each neuron in the list. If it's set to ``True`` it will record
for all the neurons in the group.
'''
M = StateMonitor(G, 'V', record=0)
'''
And then we continue as before:
'''
G.V = Vr + rand(40) * (Vt - Vr)
'''
But this time we run it for a shorter time so we can look at
the output in more detail:
'''
run(200 * msecond)
'''
Having run the simulation, we plot the results using the
``plot`` command from PyLab which has the same syntax as the Matlab
:class:`plot`` command, i.e. ``plot(xvals,yvals,...)``. The :class:`StateMonitor`
monitors the times at which it monitored a value in the
array ``M.times``, and the values in the array ``M[0]``. The notation
``M[i]`` means the array of values of the monitored state
variable for neuron ``i``.

In the following lines, we scale the times so that they're
measured in ms and the values so that they're measured in
mV. We also label the plot using PyLab's ``xlabel``, ``ylabel`` and
``title`` functions, which again mimic the Matlab equivalents.
'''
plot(M.times / ms, M[0] / mV)
xlabel('Time (in ms)')
ylabel('Membrane potential (in mV)')
title('Membrane potential for neuron 0')
show()
'''
.. image:: images/tutorials/1g.jpg

You can clearly see the leaky integration exponential decay
toward the resting potential, as well as the jumps when a
spike was received.
'''



'''
Spike threshold variability in single compartment model
-------------------------------------------------------
Figure 7 from:
Platkiewicz J, Brette R (2010). A Threshold Equation for Action Potential
Initiation. PLoS Comput Biol 6(7): e1000850. doi:10.1371/journal.pcbi.1000850

The original HH model is from Traub and Miles 1991, modified by Destexhe et al. 2001,
and we shift Na inactivation to -62 mV. This shift produces threshold variability,
but also spikes with variable shapes (unavoidable in a single-compartment model).

The script demonstrates that the spike threshold is proportional to the logarithm of h.
'''
from brian import *
from scipy import stats
from brian.library.electrophysiology import *

defaultclock.dt=0.05*ms
duration=500*ms
N=1000 # we simulate 1000 neurons to have more threshold statistics

rectify=lambda x:clip(x,0,inf)*siemens

# Biophysical parameters
#--Passive parameters
area=pi*(105*umetre)**2
C=1*uF
GL=0.0452*msiemens
EL=-80*mV
#--Active fixed parameters
celsius=36
temp=23
q10=2.3
#----Sodium channel parameters
ENa=50*mV
GNa=51.6*msiemens
vTraub_Na=-63*mV               #Traub convention  
vshift=-20*mV                #Inactivation shift (-62 mV instead of -42 mV)  
#--------activation
A_alpham=0.32/(ms*mV)                    #open (V) 
A_betam=0.28/(ms*mV)                    #close (V)
v12_alpham=13*mV                     #v1/2 for act                                                                     
v12_betam=40*mV                     #v1/2 for act
k_alpham=4*mV                       #act slope
k_betam=5*mV                       #act slope
#--------inactivation
A_alphah=0.128/ms                   #inact recov (V) 
A_betah=4/ms                    #inact (V)
v12_alphah=17*mV                    #v1/2 for inact                                                                     
v12_betah=40*mV                    #v1/2 for inact
k_alphah=18*mV                     #inact tau slope
k_betah=5*mV                      #inact tau slope
#----Potassium channel parameters
EK=-90*mV
#--------"delay-rectifier"
GK=10*msiemens
vTraub_K=-63*mV
A_alphan=0.032/(ms*mV)              #open (V) 
A_betan=0.5/ms                      #close (V)
v12_alphan=15*mV                    #v1/2 for act                                                                     
v12_betan=10*mV                     #v1/2 for act
k_alphan=5*mV                       #act slope
k_betan=40*mV                       #act slope
#--------muscarinic 
GK_m=0.5*msiemens
A_alphan_m=1e-4/(ms*mV)
A_betan_m=1e-4/(ms*mV)      
v12_alphan_m=-30*mV
v12_betan_m=-30*mV
k_alphan_m=9*mV
k_betan_m=9*mV  
# Input parameters
Ee=0*mV
Ei=-75*mV
taue=2.728*ms
taui=10.49*ms
Ge0=0.0121*usiemens*cm**2
Gi0=0.0573*usiemens*cm**2
Sigmae=0.012*usiemens*cm**2
Sigmai=0.0264*usiemens*cm**2
tadj=q10**((celsius-temp)/10)
ge0=Ge0/area
gi0=Gi0/area
sigmae=Sigmae/area
sigmai=Sigmai/area

Traubm=lambda v:v-vTraub_Na
alpham=lambda v:A_alpham*(Traubm(v)-v12_alpham)/(1-exp((v12_alpham-Traubm(v))/k_alpham)) 
betam=lambda v:-A_betam*(Traubm(v)-v12_betam)/(1-exp(-(v12_betam-Traubm(v))/k_betam))
minf=lambda v:alpham(v)/(alpham(v)+betam(v))
taum=lambda v:1/(alpham(v)+betam(v))
Shift=lambda v:Traubm(v)-vshift 
alphah=lambda v:A_alphah*exp((v12_alphah-Shift(v))/k_alphah) 
betah=lambda v:A_betah/(1+exp((v12_betah-Shift(v))/k_betah)) 
hinf=lambda v:alphah(v)/(alphah(v)+betah(v)) 
tauh=lambda v:1/(alphah(v)+betah(v))
TraubK=lambda v:v-vTraub_K 
alphan= lambda v:A_alphan*(TraubK(v)-v12_alphan)/(1-exp((v12_alphan-TraubK(v))/k_alphan)) 
betan= lambda v:A_betan*exp((v12_betan-TraubK(v))/k_betan)
ninf= lambda v:alphan(v)/(alphan(v)+betan(v)) 
taun= lambda v:1/(alphan(v)+betan(v))/tadj

eqs="""
dv/dt=(3*GNa*h*m**3*(ENa-v)+(GK*n**4+GK_m*n_m)*(EK-v)+GL*(EL-v)+I)/C : volt

# Sodium activation
m_inf=minf(v) : 1   #minf(v)
tau_m=taum(v) : second
dm/dt=(m_inf-m)/tau_m : 1

# Sodium inactivation
h_inf=hinf(v) : 1
tau_h=tauh(v) : second
dh/dt=(h_inf-h)/tau_h : 1

# Potassium : delay-rectifier
n_inf=ninf(v) : 1
tau_n=taun(v) : second
dn/dt=(n_inf-n)/tau_n : 1
gK=GK*n**4 : siemens

# Potassium : muscarinic
alphan_m=A_alphan_m*(v-v12_alphan_m)/(1-exp((v12_alphan_m-v)/k_alphan_m)) : hertz
betan_m=-A_alphan_m*(v-v12_alphan_m)/(1-exp(-(v12_alphan_m-v)/k_alphan_m)) : hertz
n_minf=alphan_m/(alphan_m+betan_m) : 1
taun_m=1/(alphan_m+betan_m)/tadj : second
dn_m/dt=(n_minf-n_m)/taun_m : 1
gK_m=GK_m*n_m : siemens

# Fluctuating synaptic conductances
I=rectify(ge)*(Ee-v)+rectify(gi)*(Ei-v) : amp
dge/dt=(1.5*ge0-ge)/taue+1.5*sigmae*(2./taue)**.5*xi : siemens
dgi/dt=(gi0-gi)/taui+2*sigmai*(2./taui)**.5*xi : siemens
gtot=GL+rectify(ge)+rectify(gi)+gK+gK_m : siemens
"""

neurons=NeuronGroup(N,model=eqs,implicit=True)
neurons.v=EL
neurons.m=minf(EL)
neurons.h=hinf(EL)
neurons.n=ninf(EL)
neurons.n_m=0
M=StateMonitor(neurons,'v',record=True)
Mh=StateMonitor(neurons,'h',record=True)

run(duration,report='text')

# Collect spike thresholds and values of h
threshold,logh=[],[]
valuesv,valuesh=array(M._values),array(Mh._values)
criterion=10*mV/ms # criterion for spike threshold
for i in range(N):
    v=valuesv[:,i]
    h=valuesh[:,i]
    onsets=spike_onsets(v,criterion=defaultclock.dt*criterion,vc=-35*mV)
    threshold.extend(v[onsets])
    logh.extend(-log(h[onsets]))#+log(g[onsets]))
threshold=array(threshold)/mV
logh=array(logh)/log(10.) # for display

# Voltage trace with spike onsets
subplot(311)
v=valuesv[:,0]
onsets=spike_onsets(v,criterion=defaultclock.dt*criterion,vc=-35*mV)
t=M.times[onsets]/ms
plot(M.times/ms,M[0]/mV,'k')
plot(t,v[onsets]/mV,'.r')
xlabel('t (ms)')
ylabel('V (mV)')

# Distribution of Vm and spike onsets
subplot(312)
hist(threshold,30,normed=1,histtype='stepfilled',alpha=0.6,facecolor='r')
hist(valuesv.flatten()/mV,100,normed=1,histtype='stepfilled',alpha=0.6,facecolor='k')
xlabel('V (mV)')
xlim(-80,-40)

# Relationship between h and spike threshold
subplot(313)
slope,intercept=3.1,-54. # linear regression for the prediction of threshold
p1,p2=min(logh),max(logh)
plot(logh[:len(logh)/10],threshold[:len(logh)/10],'.k') # don't show everything
plot([p1,p2],intercept+slope*array([p1,p2])*log(10.),'r')
xlabel('h')
ylabel('Threshold (mV)')
ylim(-55,-40)
xticks([0,-log(5e-1)/log(10),1,-log(5e-2)/log(10)],[1,5e-1,1e-1,5e-2])
xlim(0,1.5)

show()



# Reward - based STDP based on code from nemo136
# posted on Brian Support Google Group

from brian import *

#---------------------------------------------------
# Variable declarations
#---------------------------------------------------


# Model variables and equations
dt = 1 * ms
tau = 10 * ms
Vr = -70 * mV
Vl = -55 * mV
lif_eqs = '''
dVm/dt=(Vl-Vm)/tau : volt
K : volt
'''

# Basic STDP variables
tau_pre = 10 * ms
tau_post = 10 * ms
gmax = 1
dA_pre = .001
dA_post = -dA_pre * tau_pre / tau_post * 2.5


#-------------------------------------------------------
# Set up neuron groups
#-------------------------------------------------------

# Input group of Poisson neurons
groupi = PoissonGroup(100, rates=linspace(0.1 * Hz, 10 * Hz, 100))

# Main neuron population, split into two subgroups

groupa = NeuronGroup(100, model=lif_eqs, threshold= -55 * mV,
                     reset=Vr, max_delay=6 * ms,)
suba = groupa[0:50]
subb = groupa[50:100]

#-------------------------------------------------------
# Create connections
#-------------------------------------------------------

# Connect input to group a - this inputs to Voltage in main eqs

input_signal = Connection(groupi, suba, 'Vm', weight=0.2 * mV,
                          sparseness=0.6, delay=4 * ms)

# Connect group a to b - this inputs to Voltage in main eqs

synapses = Connection(suba, subb, 'Vm', weight=0.1 * mV,
                      sparseness=0.7, delay=3 * ms, structure='sparse', column_access=True)

# Make trace connection which is a copy of synapses connection

trace = Connection(suba, subb, 'K', weight=synapses.W)

trace.compress()

T_matrix = trace.W

#-------------------------------------------------------
# Set up STDP on trace connection to collect the weight
# changes rather than apply them immediately
#-------------------------------------------------------

eqs_stdp = '''
dA_pre/dt=-A_pre/tau_pre : 1
dA_post/dt=-A_post/tau_post : 1
'''
stdp = STDP(trace, eqs=eqs_stdp, pre='A_pre+=dA_pre;w+=A_post',
                  post='A_post+=dA_post;w+=A_pre', wmax=gmax)


def update_trace(spikes):
        if len(spikes) > 0:
            # Apply reward to STDP weights which have been
            # collected in trace
            for j in spikes:
                T_matrix[:, j] *= 0.05
#            for i in xrange(len(trace.W.get_col(0))):
#                for j in spikes:
#                    if not isinstance(T_matrix[i,j],int):
#                        T_matrix[i,j] *= 0.05


#---------------------------------------------------------
# Create monitors
#---------------------------------------------------------


# Custom SpikeMonitor for applying reward

Sa1 = SpikeMonitor(suba, function=update_trace)


# Set clock
defaultclock.dt = 1.0 * ms

# Initialise the Voltage to resting voltage, Vr

groupa.Vm = Vr

# Run

run(500 * msecond)



'''
Synfire chains
--------------
M. Diesmann et al. (1999). Stable propagation of synchronous spiking in cortical
neural networks. Nature 402, 529-533.
'''
from brian import *
# Neuron model parameters
Vr = -70 * mV
Vt = -55 * mV
taum = 10 * ms
taupsp = 0.325 * ms
weight = 4.86 * mV
# Neuron model
eqs = Equations('''
dV/dt=(-(V-Vr)+x)*(1./taum) : volt
dx/dt=(-x+y)*(1./taupsp) : volt
dy/dt=-y*(1./taupsp)+25.27*mV/ms+\
    (39.24*mV/ms**0.5)*xi : volt
''')
# Neuron groups
P = NeuronGroup(N=1000, model=eqs,
    threshold=Vt, reset=Vr, refractory=1 * ms)
Pinput = PulsePacket(t=50 * ms, n=85, sigma=1 * ms)
# The network structure
Pgp = [ P.subgroup(100) for i in range(10)]
C = Connection(P, P, 'y')
for i in range(9):
    C.connect_full(Pgp[i], Pgp[i + 1], weight)
Cinput = Connection(Pinput, Pgp[0], 'y')
Cinput.connect_full(weight=weight)
# Record the spikes
Mgp = [SpikeMonitor(p) for p in Pgp]
Minput = SpikeMonitor(Pinput)
monitors = [Minput] + Mgp
# Setup the network, and run it
P.V = Vr + rand(len(P)) * (Vt - Vr)
run(100 * ms)
# Plot result
raster_plot(showgrouplines=True, *monitors)
show()



from brian import *

bof = 3.0
tau = 10 * ms

eq_base = Equations('''
w = V*bof : 1
''')

eq_extra_1 = Equations('''
dV/dt = (w-V)/tau : 1
''')

eq_extra_2 = Equations('''
dV/dt = (2*w-V)/tau : 1
''')

eq1 = eq_extra_1 + eq_base
eq2 = eq_extra_2 + eq_base

print eq1._namespace['w'].keys()

eq1.prepare()

print eq1._namespace['w'].keys()

eq2.prepare()

#exit()

G1 = NeuronGroup(1, eq1)
G2 = NeuronGroup(1, eq2)

run(1 * ms)

print G1.V
print G1.w
print G2.V
print G2.w



import brian_no_units
from brian import *
from itertools import count
from scipy.signal import lfilter
from scipy import linalg
from numpy.linalg import inv, matrix_power
import numpy
import time

def get_linear_equations(eqs):
    '''
    Returns the matrices A and C for the linear model dX/dt = M(X-B),
    where eqs is an Equations object. 
    '''
    # Otherwise assumes it is given in functional form
    n = len(eqs._diffeq_names) # number of state variables
    dynamicvars = eqs._diffeq_names
    
    #print eqs._diffeq_names

    # Calculate B
    AB = zeros((n, 1))
    d = dict.fromkeys(dynamicvars)

    #print d

    for j in range(n):
        d[dynamicvars[j]] = 0. * eqs._units[dynamicvars[j]]
    for var, i in zip(dynamicvars, count()):
        #print i, var
        AB[i] = -eqs.apply(var, d)
    # Calculate A
    M = zeros((n, n))
    for i in range(n):
        for j in range(n):
            d[dynamicvars[j]] = 0. * eqs._units[dynamicvars[j]]
        if isinstance(eqs._units[dynamicvars[i]], Quantity):
            d[dynamicvars[i]] = Quantity.with_dimensions(1., eqs._units[dynamicvars[i]].get_dimensions())
        else:
            d[dynamicvars[i]] = 1.
        for var, j in zip(dynamicvars, count()):
            M[j, i] = eqs.apply(var, d) + AB[j]
    #M-=eye(n)*1e-10 # quick dirty fix for problem of constant derivatives; dimension = Hz
    #B=linalg.lstsq(M,AB)[0] # We use this instead of solve in case M is degenerate
    
    try:
        B = linalg.solve(M, AB) # We use this instead of solve in case M is degenerate
    except:
        B = None

    return M, B

def compute_filter(A, row=0):
    d = len(A)

    # compute a
    a = poly(A)  # directly vector a of the filter, a[0]=1

    # compute b recursively
    b = zeros(d+1)
    T = eye(d)
    b[0] = T[row, 0]
    for i in range(1, d+1):
        T = a[i]*eye(d) + dot(A, T)
        b[i] = T[row, 0]

    return b, a

def simulate(eqs, I, dt, row=0):
    """
    I must be normalized (I*Re/taue for example)
    """
    M, B = get_linear_equations(eqs)
    A = linalg.expm(M * dt)
    b, a = compute_filter(A, row=row)
    y = lfilter(b, a, I*dt) + B[row]
    return y

if __name__ == '__main__':
    R = 500*Mohm
    Re = 400*Mohm
    tau = 10*ms
    taue = 1.0*ms
    Vr = -70*mV
    dt = .1*ms

    # +Re*I/taue
    eqs = Equations("""
        dV/dt=Re*(-Iinj)/taue : volt
        dV0/dt=(R*Iinj-V0+Vr)/tau : volt
        Iinj=(V-V0)/R : amp
        """)
    eqs.prepare()


    

    Is = numpy.load("current1.npy")[:50000]
    eqs2 = Equations("""
        dV/dt=Re*(I-Iinj)/taue : volt
        dV0/dt=(R*Iinj-V0+Vr)/tau : volt
        Iinj=(V-V0)/R : amp
        I : amp
    """)

    G = NeuronGroup(1, eqs2)
    G.V = Vr
    G.I = TimedArray(Is)
    stm = StateMonitor(G, 'V', record=True)
    t0 = time.clock()
    run(len(Is)*dt)
    t1 = time.clock()-t0
    y0 = stm.values[0]

    t0 = time.clock()
    y = simulate(eqs, Is * Re/taue, dt, row=0)
    t2 = time.clock()-t0

    print t1, t2, t1/t2
    print max(abs(y-y0))

    subplot(211)
    plot(Is)
    subplot(212)
    plot(y0)
    plot(y)
    show()



"""
Naundorf et al. (2006)
Unique features...
Nature.

Doesn't work... typos?
"""
from brian import *

mS=msiemens

v12A=-35*mV
kA=6*mV
invtauA=1/(0.1*ms)
invtauI=1/(0.5*ms)
v12C=80*mV # -80 mV?
kC=4*mV
invtauC=1/(30*ms)
C=1*uF
gL=2*mS
VL=-80*mV
gNa=68.4*mS
VNa=50*mV
I0=0*uA
sigma=12*uA
K=1000
J=3.2*mV
tau=50*ms

eqs="""
dv/dt=(gL*(VL-v)+gNa*o*(VNa-v)+I0+sigma*z)/C : volt
dz/dt=-z/tau+tau**-.5*xi : 1 # noisy current
do/dt=alphaA*(h-o)-(invtauI+betaA)*o : 1 # open channels
dh/dt=alphaC*(1-h)-betaC*(h-o)-invtauI*o : 1 # available channels
# Channel kinetics
alphaA=invtauA/(1+exp(-(v+K*J*o-v12A)/kA)) : Hz
betaA=invtauA/(1+exp((v+K*J*o-v12A)/kA)) : Hz
alphaC=invtauC/(1+exp(-(v-v12C)/kC)) : Hz # I corrected * to /
betaC=invtauC/(1+exp((v-v12C)/kC)) : Hz # idem
"""

neuron=NeuronGroup(1,eqs,threshold=0*mV,reset=VL)
neuron.v=VL
neuron.o=0
neuron.h=1

M=StateMonitor(neuron,'v',record=0)

run(2*second,report='text')

plot(M.times/ms,M[0]/mV)

show()



import pycuda.autoinit as autoinit
import pycuda.driver as drv
from pycuda.compiler import SourceModule
from pycuda.gpuarray import GPUArray
from pycuda import gpuarray
import bisect
import numpy, pylab, time
from itertools import repeat
import cProfile as profile
import pstats
import brian

N = 10000
blocksize = 512
N = int(numpy.ceil(1. * N / blocksize) * blocksize)
duration = 10000
record = False
doprofile = False
sparseness = 80. / N

if drv.get_version() == (2, 0, 0): # cuda version
    precision = 'float'
elif drv.get_version() > (2, 0, 0):
    precision = 'double'
else:
    raise Exception, "CUDA 2.0 required"

if precision == 'double':
    mydtype = numpy.float64
else:
    mydtype = numpy.float32

block = (blocksize, 1, 1)
grid = (N / blocksize, 1)
Ne = int(N * 0.8)
Ni = N - Ne

shared_memory_size = 2 ** 14 # 16k
matrix_split_size = blocksize * (shared_memory_size // mydtype().nbytes // blocksize)
print 'Matrix split size:', matrix_split_size

mod = SourceModule("""
__global__ void stateupdate_and_threshold(SCALAR *V_arr, SCALAR *ge_arr, SCALAR *gi_arr, int *spikes, bool *spiked, unsigned int *global_j, int N)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    SCALAR V = V_arr[i];
    SCALAR ge = ge_arr[i];
    SCALAR gi = gi_arr[i]; 
    SCALAR V__tmp = (ge+gi-(V+0.049))/0.02;
    SCALAR ge__tmp = -ge/0.005;
    SCALAR gi__tmp = -gi/0.01;
    V += 0.0001*V__tmp;
    ge += 0.0001*ge__tmp;
    gi += 0.0001*gi__tmp;
    V_arr[i] = V;
    ge_arr[i] = ge;
    gi_arr[i] = gi;
    bool this_spiked = V>-0.05; 
    spiked[i] = this_spiked;
    if(this_spiked)
        spikes[atomicInc(global_j, N)] = i;
}

__global__ void reset(SCALAR *V, bool *spiked)
{
    int i = blockIdx.x * blockDim.x + threadIdx.x;
    bool has_spiked = spiked[i];
    //V[i] = (V[i]*!has_spiked)+(-0.06)*has_spiked; // i.e. V[i]=-0.06 if spiked[i] - but this operation requires two global mem accesses
    if(has_spiked)
      V[i] = -0.06;
}
""".replace('SCALAR', precision))
stateupdate_and_threshold = mod.get_function("stateupdate_and_threshold")
reset = mod.get_function("reset")

V = gpuarray.to_gpu(numpy.array(numpy.random.rand(N) * 0.01 - 0.06, dtype=mydtype))
ge = gpuarray.to_gpu(numpy.zeros(N, dtype=mydtype))
gi = gpuarray.to_gpu(numpy.zeros(N, dtype=mydtype))

we = 0.00162
wi = -0.009
P = brian.NeuronGroup(N, model='V:1')
Pe = P.subgroup(Ne)
Pi = P.subgroup(Ni)
Ce = brian.Connection(Pe, P, sparseness=sparseness, weight=we).W.connection_matrix()
Ci = brian.Connection(Pi, P, sparseness=sparseness, weight=wi).W.connection_matrix()
Ce.alldata = numpy.array(Ce.alldata, dtype=mydtype)
Ci.alldata = numpy.array(Ci.alldata, dtype=mydtype)
Ce_alldata = drv.mem_alloc(Ce.alldata.nbytes)
drv.memcpy_htod(Ce_alldata, Ce.alldata)
Ce_allj = drv.mem_alloc(Ce.allj.nbytes)
drv.memcpy_htod(Ce_allj, Ce.allj)
Ce_rowind = drv.mem_alloc(Ce.rowind.nbytes)
drv.memcpy_htod(Ce_rowind, Ce.rowind)
Ci_alldata = drv.mem_alloc(Ci.alldata.nbytes)
drv.memcpy_htod(Ci_alldata, Ci.alldata)
Ci_allj = drv.mem_alloc(Ci.allj.nbytes)
drv.memcpy_htod(Ci_allj, Ci.allj)
Ci_rowind = drv.mem_alloc(Ci.rowind.nbytes)
drv.memcpy_htod(Ci_rowind, Ci.rowind)



'''
Example of a simple auditory nerve fibre model with Brian hears.
'''
from brian import *
from brian.hears import *

sound1 = tone(1*kHz, .1*second)
sound2 = whitenoise(.1*second)

sound = sound1+sound2
sound = sound.ramp()

cf = erbspace(20*Hz, 20*kHz, 3000)
cochlea = Gammatone(sound, cf)

# Half-wave rectification and compression [x]^(1/3)
ihc = FunctionFilterbank(cochlea, lambda x: 3*clip(x, 0, Inf)**(1.0/3.0))

# Leaky integrate-and-fire model with noise and refractoriness
eqs = '''
dv/dt = (I-v)/(1*ms)+0.2*xi*(2/(1*ms))**.5 : 1
I : 1
'''
anf = FilterbankGroup(ihc, 'I', eqs, reset=0, threshold=1, refractory=5*ms)

M = SpikeMonitor(anf)
run(sound.duration)
raster_plot(M)
show()



from brian import *

__all__ = ['brian_sample_run']

def brian_sample_run():
    '''
    A simple Brian script for users to test they have installed Brian correctly.
    '''
    reinit_default_clock()
    clear(True)
    eqs = '''
    dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
    dge/dt = -ge/(5*ms) : volt
    dgi/dt = -gi/(10*ms) : volt
    '''
    P = NeuronGroup(4000, eqs, threshold= -50 * mV, reset= -60 * mV)
    P.v = -60 * mV + 10 * mV * rand(len(P))
    Pe = P.subgroup(3200)
    Pi = P.subgroup(800)
    Ce = Connection(Pe, P, 'ge')
    Ci = Connection(Pi, P, 'gi')
    Ce.connect_random(Pe, P, 0.02, weight=1.4 * mV)
    Ci.connect_random(Pi, P, 0.02, weight= -9 * mV)
    M = SpikeMonitor(P)
    run(1000 * ms)
    raster_plot(M)
    print 'Brian sample run finished OK!'
    show()

if __name__ == '__main__':
    brian_sample_run()



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
'''
Network class
'''
__all__ = ['Network', 'MagicNetwork', 'NetworkOperation', 'network_operation', 'run',
           'reinit', 'stop', 'clear', 'forget', 'recall']

from Queue import Queue
from connections import *
from neurongroup import NeuronGroup
from clock import guess_clock, Clock
import magic
from inspect import *
from operator import isSequenceType
import types
from itertools import chain
from collections import defaultdict
import copy
from base import *
from units import second
import time
from utils.progressreporting import *
from globalprefs import *
import gc
import heapq

globally_stopped = False


class Network(object):
    '''
    Contains simulation objects and runs simulations
    
    **Initialised as:** ::
    
        Network(...)
    
    with ``...`` any collection of objects that should be added to the :class:`Network`.
    You can also pass lists of objects, lists of lists of objects, etc. Objects
    that need to passed to the :class:`Network` object are:
    
    * :class:`NeuronGroup` and anything derived from it such as :class:`PoissonGroup`.
    * :class:`Connection` and anything derived from it.
    * Any monitor such as :class:`SpikeMonitor` or :class:`StateMonitor`.
    * Any network operation defined with the :func:`network_operation` decorator.
    
    Models, equations, etc. do not need to be passed to the :class:`Network` object. 
    
    The most important method is the ``run(duration)`` method which runs the simulation
    for the given length of time (see below for details about what happens when you
    do this).
    
    **Example usage:** ::
    
        G = NeuronGroup(...)
        C = Connection(...)
        net = Network(G,C)
        net.run(1*second)
    
    **Methods**
    
    ``add(...)``
        Add additional objects after initialisation, works the same way
        as initialisation.
    ``run(duration[, report[, report_period]])``
        Runs the network for the given duration. See below for details about
        what happens when you do this. See documentation for :func:`run` for
        an explanation of the ``report`` and ``report_period`` keywords.
    ``reinit(states=True)``
        Reinitialises the network, runs each object's ``reinit()`` and each
        clock's ``reinit()`` method (resetting them to 0). If ``states=False``
        then it will not reinitialise the :class:`NeuronGroup` state variables.
    ``stop()``
        Can be called from a :func:`network_operation` for example to stop the
        network from running.
    ``__len__()``
        Returns the number of neurons in the network.
    ``__call__(obj)``
        Similar to ``add``, but you can only pass one object and that
        object is returned. You would only need this in obscure
        circumstances where objects needed to be added to the network
        but were either not stored elsewhere or were stored in a way
        that made them difficult to extract, for example below the
        NeuronGroup object is only added to the network if certain
        conditions hold::
        
            net = Network(...)
            if some_condition:
                x = net(NeuronGroup(...))
    
    **What happens when you run**
    
    For an overview, see the Concepts chapter of the main documentation.
    
    When you run the network, the first thing that happens is that it
    checks if it has been prepared and calls the ``prepare()`` method
    if not. This just does various housekeeping tasks and optimisations
    to make the simulation run faster. Also, an update schedule is
    built at this point (see below).
    
    Now the ``update()`` method is repeatedly called until every clock
    has run for the given length of time. After each call of the
    ``update()`` method, the clock is advanced by one tick, and if
    multiple clocks are being used, the next clock is determined (this
    is the clock whose value of ``t`` is minimal amongst all the clocks).
    For example, if you had two clocks in operation, say ``clock1`` with
    ``dt=3*ms`` and ``clock2`` with ``dt=5*ms`` then this will happen:
    
    1. ``update()`` for ``clock1``, tick ``clock1`` to ``t=3*ms``, next
       clock is ``clock2`` with ``t=0*ms``.
    2. ``update()`` for ``clock2``, tick ``clock2`` to ``t=5*ms``, next
       clock is ``clock1`` with ``t=3*ms``.
    3. ``update()`` for ``clock1``, tick ``clock1`` to ``t=6*ms``, next
       clock is ``clock2`` with ``t=5*ms``.
    4. ``update()`` for ``clock2``, tick ``clock2`` to ``t=10*ms``, next
       clock is ``clock1`` with ``t=6*ms``.
    5. ``update()`` for ``clock1``, tick ``clock1`` to ``t=9*ms``, next
       clock is ``clock1`` with ``t=9*ms``.
    6. ``update()`` for ``clock1``, tick ``clock1`` to ``t=12*ms``, next
       clock is ``clock2`` with ``t=10*ms``. etc.
    
    The ``update()`` method simply runs each operation in the current clock's
    update schedule. See below for details on the update schedule.
    
    **Update schedules**
    
    An update schedule is the sequence of operations that are
    called for each ``update()`` step. The standard update schedule is:
    
    *  Network operations with ``when = 'start'``
    *  Network operations with ``when = 'before_groups'``
    *  Call ``update()`` method for each :class:`NeuronGroup`, this typically
       performs an integration time step for the differential equations
       defining the neuron model.
    *  Network operations with ``when = 'after_groups'``
    *  Network operations with ``when = 'middle'``
    *  Network operations with ``when = 'before_connections'``
    *  Call ``do_propagate()`` method for each :class:`Connection`, this
       typically adds a value to the target state variable of each neuron
       that a neuron that has fired is connected to. See Tutorial 2: Connections for
       a more detailed explanation of this.
    *  Network operations with ``when = 'after_connections'``
    *  Network operations with ``when = 'before_resets'``
    *  Call ``reset()`` method for each :class:`NeuronGroup`, typically resets a
       given state variable to a given reset value for each neuron that fired
       in this update step.
    *  Network operations with ``when = 'after_resets'``
    *  Network operations with ``when = 'end'``
    
    There is one predefined alternative schedule, which you can choose by calling
    the ``update_schedule_groups_resets_connections()`` method before running the
    network for the first time. As the name suggests, the reset operations are
    done before connections (and the appropriately named network operations are
    called relative to this rearrangement). You can also define your own update
    schedule with the ``set_update_schedule`` method (see that method's API documentation for
    details). This might be useful for example if you have a sequence of network
    operations which need to be run in a given order.    
    '''

    operations = property(fget=lambda self:self._all_operations)

    def __init__(self, *args, **kwds):
        self.clock = None # Initialized later
        self.groups = []
        self.connections = []
        # The following dict keeps a copy of which operations are in which slot
        self._operations_dict = defaultdict(list)
        self._all_operations = []
        self.update_schedule_standard()
        self.prepared = False
        for o in chain(args, kwds.itervalues()):
            self.add(o)

    def add(self, *objs):
        """
        Add an object or container of objects to the network
        """
        for obj in objs:
            if isinstance(obj, NeuronGroup):
                if obj not in self.groups:
                    self.groups.append(obj)
            elif isinstance(obj, Connection):
                if obj not in self.connections:
                    self.connections.append(obj)
            elif isinstance(obj, NetworkOperation):
                if obj not in self._all_operations:
                    self._operations_dict[obj.when].append(obj)
                    self._all_operations.append(obj)
            elif isSequenceType(obj):
                for o in obj:
                    self.add(o)
            else:
                raise TypeError('Only the following types of objects can be added to a network: NeuronGroup, Connection or NetworkOperation')

            try:
                gco = obj.contained_objects
                if gco is not None:
                    self.add(gco)
            except AttributeError:
                pass

    def __call__(self, obj):
        """
        Add an object to the network and return it
        """
        self.add(obj)
        return obj

    def reinit(self, states=True):
        '''
        Resets the objects and clocks. If ``states=False`` it will not reinit
        the state variables.
        '''
        objs = self.groups + self.connections + self.operations
        if self.clock is not None:
            objs.append(self.clock)
        else:
            guess_clock(None).reinit()
        if hasattr(self, 'clocks'):
            objs.extend(self.clocks)
        for P in objs:
            if hasattr(P, 'reinit'):
                if isinstance(P, NeuronGroup):
                    try:
                        P.reinit(states=states)
                    except TypeError:
                        P.reinit()
                else:
                    P.reinit()

    def prepare(self):
        '''
        Prepares the network for simulation:
        + Checks the clocks of the neuron groups
        + Gather connections with identical subgroups
        + Compresses the connection matrices for faster simulation
        Calling this function is not mandatory but speeds up the simulation.
        '''
        # Set the clock
        if self.same_clocks():
            self.set_clock()
        else:
            self.set_many_clocks()

        # Gather connections with identical subgroups
        # 'subgroups' maps subgroups to connections (initialize with immutable object (not [])!)
        subgroups = dict.fromkeys([(C.source, C.delay) for C in self.connections], None)
        for C in self.connections:
            if subgroups[(C.source, C.delay)] == None:
                subgroups[(C.source, C.delay)] = [C]
            else:
                subgroups[(C.source, C.delay)].append(C)
        self.connections = subgroups.values()
        cons = self.connections # just for readability
        for i in range(len(cons)):
            if len(cons[i]) > 1: # at least 2 connections with the same subgroup
                cons[i] = MultiConnection(cons[i][0].source, cons[i])
            else:
                cons[i] = cons[i][0]

        # Compress connections
        for C in self.connections:
            C.compress()

        # Experimental support for new propagation code
        if get_global_preference('usenewpropagate') and get_global_preference('useweave'):
            from experimental.new_c_propagate import make_new_connection
            for C in self.connections:
                make_new_connection(C)

        # build operations list for each clock
        self._build_update_schedule()

        self.prepared = True

    def update_schedule_standard(self):
        self._schedule = ['ops start',
                          'ops before_groups',
                          'groups',
                          'ops after_groups',
                          'ops middle',
                          'ops before_connections',
                          'connections',
                          'ops after_connections',
                          'ops before_resets',
                          'resets',
                          'ops after_resets',
                          'ops end'
                          ]
        self._build_update_schedule()

    def update_schedule_groups_resets_connections(self):
        self._schedule = ['ops start',
                          'ops before_groups',
                          'groups',
                          'ops after_groups',
                          'ops middle',
                          'ops before_resets',
                          'resets',
                          'ops after_resets',
                          'ops before_connections',
                          'connections',
                          'ops after_connections',
                          'ops end'
                          ]
        self._build_update_schedule()

    def set_update_schedule(self, schedule):
        """
        Defines a custom update schedule
        
        A custom update schedule is a list of schedule items. Each update
        step of the network, the schedule items will be run in turn. A
        schedule item can be defined as a string or tuple. The following
        string definitions are possible:
        
        'groups'
            Calls the 'update' function of each group in turn, this is
            typically the integration step of the simulation.
        'connections'
            Calls the 'do_propagate' function of each connection in
            turn, this is typically propagating spikes forward (and
            backward in the case of STDP).
        'resets'
            Calls the 'reset' function of each group in turn.
        'ops '+name
            Calls each operation in turn whose 'when' parameter is
            set to 'name'. The standard set of 'when' names is
            start, before_groups, after_groups, before_resets,
            after_resets, before_connections, after_connections,
            end, but you can use any you like.
        
        If a tuple is provided, it should be of the form:
        
            (objset, func, allclocks)
        
        with:
        
        objset
            a list of objects to be processed
        func
            Either None or a string. In the case of none, each
            object in objset must be callable and will be called.
            In the case of a string, obj.func will be called for
            each obj in objset.
        allclocks
            Either True or False. If it's set to True, then the
            object will be placed in the update schedule of
            every clock in the network. If False, it will be
            placed in the update schedule only of the clock
            obj.clock.
        """
        self._schedule = schedule
        self._build_update_schedule()

    def _build_update_schedule(self):
        '''
        Defines what the update step does
        
        For each clock we build a list self._update_schedule[id(clock)]
        of functions which are called at the update step if that clock
        is active. This is generic and works for single or multiple
        clocks.
        
        See documentation for set_update_schedule for an explanation of
        the self._schedule object. 
        '''
        self._update_schedule = defaultdict(list)
        if hasattr(self, 'clocks'):
            clocks = self.clocks
        else:
            clocks = [self.clock]
        clockset = clocks
        for item in self._schedule:
            # we define some simple names for common schedule items
            if isinstance(item, str):
                if item == 'groups':
                    objset = self.groups
                    objfun = 'update'
                    allclocks = False
                elif item == 'resets':
                    objset = self.groups
                    objfun = 'reset'
                    allclocks = False
                elif item == 'connections':
                    objset = self.connections
                    objfun = 'do_propagate'
                    allclocks = False
                    # Connections do not define their own clock, but they should
                    # be updated on the schedule of their source group
                    for obj in objset:
                        obj.clock = obj.source.clock
                elif len(item) > 4 and item[0:3] == 'ops': # the item is of the forms 'ops when'
                    objset = self._operations_dict[item[4:]]
                    objfun = None
                    allclocks = False
            else:
                # we allow the more general form of usage as well
                objset, objfun, allclocks = item
            for obj in objset:
                if objfun is None:
                    f = obj
                else:
                    f = getattr(obj, objfun)
                if not allclocks:
                    useclockset = [obj.clock]
                else:
                    useclockset = clockset
                for clock in useclockset:
                    self._update_schedule[id(clock)].append(f)

    def update(self):
        for f in self._update_schedule[id(self.clock)]:
            f()

    """
    def update_threaded(self,queue):
        '''
        EXPERIMENTAL (not useful for the moment)
        Parallel update of the network (using threads).
        '''
        # Update groups: one group = one thread
        for P in self.groups:
            queue.put(P)
        queue.join() # Wait until job is done

        # The following is done serially
        # Propagate spikes
        for C in self.connections:
            C.propagate(C.source.get_spikes(C.delay))
            
        # Miscellanous operations
        for op in self.operations:
            op()
    """

    def run(self, duration, threads=1, report=None, report_period=10 * second):
        '''
        Runs the simulation for the given duration.
        '''
        global globally_stopped
        self.stopped = False
        globally_stopped = False
        if not self.prepared:
            self.prepare()
        self.clock.set_duration(duration)
        try:
            for c in self.clocks:
                c.set_duration(duration)
        except AttributeError:
            pass
        if report is not None:
            start_time = time.time()
            if not isinstance(report, ProgressReporter):
                report = ProgressReporter(report, report_period)
                next_report_time = start_time + float(report_period)
            else:
                report_period = report.period
                next_report_time = report.next_report_time

        if self.clock.still_running() and not self.stopped and not globally_stopped:
            not_same_clocks = not self.same_clocks()
            clk = self.clock
            while clk.still_running() and not self.stopped and not globally_stopped:
                if report is not None:
                    cur_time = time.time()
                    if cur_time > next_report_time:
                        next_report_time = cur_time + float(report_period)
                        report.update((self.clock.t - self.clock.start) / duration)
                self.update()
                clk.tick()
                if not_same_clocks:
                    # Find the next clock to update
                    #self.clock = min([(clock.t, id(clock), clock) for clock in self.clocks])[2]
                    clk = self.clock = min(self.clocks)
                    #heapq.heappush(self.clocks, self.clock)
                    #clk = self.clock = heapq.heappop(self.clocks)
                    #clk = self.clock = heapq.heappushpop(self.clocks, self.clock)
                    #if not clk<self.clocks[0]:
                    #    clk = self.clock = heapq.heapreplace(self.clocks, self.clock)
        if report is not None:
            report.update(1.0)

    def stop(self):
        '''
        Stops the network from running, this is reset the next time ``run()`` is called.
        '''
        self.stopped = True

    def same_clocks(self):
        '''
        Returns True if the clocks of all groups and operations are the same.
        '''
        groups_and_operations=self.groups + self.operations
        if len(groups_and_operations)>0:
            clock = groups_and_operations[0].clock
            return all([obj.clock == clock for obj in groups_and_operations])
        else:
            return True

    def set_clock(self):
        '''
        Sets the clock and checks that clocks of all groups are synchronized.
        '''
        if self.same_clocks():
            groups_and_operations=self.groups + self.operations
            if len(groups_and_operations)>0:
                self.clock = groups_and_operations[0].clock
            else:
                self.clock = guess_clock()
        else:
            raise TypeError, 'Clocks are not synchronized!' # other error type?

    def set_many_clocks(self):
        '''
        Sets a list of clocks.
        self.clock points to the current clock between considered.
        '''
        self.clocks = list(set([obj.clock for obj in self.groups + self.operations]))
        #self.clocks.sort(key=lambda c:-c._dt)
        #heapq.heapify(self.clocks)
        #self.clock = min([(clock.t, clock) for clock in self.clocks])[1]
        self.clock = min(self.clocks)
        #self.clock = heapq.heappop(self.clocks)

    def __len__(self):
        '''
        Number of neurons in the network
        '''
        n = 0
        for P in self.groups:
            n += len(P) # use compact iterator function?
        return n

    def __repr__(self):
        return 'Network of' + str(len(self)) + 'neurons'

    # TODO: obscure custom update schedules might still lead to unpicklable Network object
    def __reduce__(self):
        # This code might need some explanation:
        #
        # The problem with pickling the Network object is that you cannot pickle
        # 'instance methods', that is a copy of a method of an instance. The
        # Network object does this because the _update_schedule attribute stores
        # a copy of all the functions that need to be called each time step, and
        # these are all instance methods (of NeuronGroup, Reset, etc.). So, we
        # solve the problem by deleting this attribute at pickling time, and then
        # rebuilding it at unpickling time. The function unpickle_network defined
        # below does the unpickling.
        #
        # We basically want to make a copy of the current object, and delete the
        # update schedule from it, and then pickle that. Some weird recursive
        # stuff happens if you try to do this in the obvious way, so we take the
        # seemingly mad step of converting the object to a general 'heap' class
        # (that is, a new-style class with no methods or anything, in this case
        # the NetworkNoMethods class defined below), do all our operations on
        # this, store a copy of the actual class of the object (which may not be
        # Network for derived classes), work with this, and then restore
        # everything back to the way it was when everything is done.
        #
        oldclass = self.__class__ # class may be derived from Network
        self.__class__ = NetworkNoMethods # stops recursion in copy.copy
        net = copy.copy(self) # we make a copy because after returning from this function we can't restore the class
        self.__class__ = oldclass # restore the class of the original, which is now back in its original state
        net._update_schedule = None # remove the problematic element from the copy
        return (unpickle_network, (oldclass, net)) # the unpickle_network function called with arguments oldclass, net restores it as it was

# This class just used as a general 'heap' class - has no methods but can have attributes
class NetworkNoMethods(object):
    pass

def unpickle_network(oldclass, net):
    # See Network.__reduce__ for an explanation, basically the _update_schedule
    # cannot be pickled because it contains instance methods, but it can just be
    # rebuilt.
    net.__class__ = oldclass
    net._build_update_schedule()
    return net


class NetworkOperation(magic.InstanceTracker, ObjectContainer):
    """Callable class for operations that should be called every update step
    
    Typically, you should just use the :func:`network_operation` decorator, but if you
    can't for whatever reason, use this. Note: current implementation only works for
    functions, not any callable object.
    
    **Initialisation:** ::
    
        NetworkOperation(function[,clock])

    If your function takes an argument, the clock will be passed
    as that argument.
    """
    def __init__(self, function=None, clock=None, when='end'):
        self.clock = guess_clock(clock)
        self.when = when
        self.function = function
        if hasattr(function, 'func_code'):
            self._has_arg = (self.function.func_code.co_argcount==1)

    def __call__(self):
        if not hasattr(self, 'function') or self.function is None:
            return
        if self._has_arg:
            self.function(self.clock)
        else:
            self.function()


def network_operation(*args, **kwds):
    """Decorator to make a function into a :class:`NetworkOperation`
    
    A :class:`NetworkOperation` is a callable class which is called every
    time step by the :class:`Network` ``run`` method. Sometimes it is useful
    to just define a function which is to be run every update step. This
    decorator can be used to turn a function into a :class:`NetworkOperation`
    to be added to a :class:`Network` object.
    
    **Example usages**
    
    Operation doesn't need a clock::
    
        @network_operation
        def f():
            ...
        
    Automagically detect clock::
    
        @network_operation
        def f(clock):
            ...
    
    Specify a clock::
    
        @network_operation(specifiedclock)
        def f(clock):
            ...
        
    Specify when the network operation is run (default is ``'end'``)::
    
        @network_operation(when='start')
        def f():
            ...
    
    Then add to a network as follows::
    
        net = Network(f,...)
    """
    # Notes on this decorator:
    # Normally, a decorator comes in two types, with or without arguments. If
    # it has no arguments, e.g.
    #   @decorator
    #   def f():
    #      ...
    # then the decorator function is defined with an argument, and that
    # argument is the function f. In this case, the decorator function
    # returns a new function in place of f.
    #
    # However, you can also define:
    #   @decorator(arg)
    #   def f():
    #      ...
    # in which case the argument to the decorator function is arg, and the
    # decorator function returns a 'function factory', that is a callable
    # object that takes a function as argument and returns a new function.
    #
    # It might be clearer just to note that the first form above is equivalent
    # to:
    #   f = decorator(f)
    # and the second to:
    #   f = decorator(arg)(f)
    #
    # In this case, we're allowing the decorator to be called either with or
    # without an argument, so we have to look at the arguments and determine
    # if it's a function argument (in which case we do the first case above),
    # or if the arguments are arguments to the decorator, in which case we
    # do the second case above.
    #
    # Here, the 'function factory' is the locally defined class
    # do_network_operation, which is a callable object that takes a function
    # as argument and returns a NetworkOperation object.
    class do_network_operation(object):
        def __init__(self, clock=None, when='end'):
            self.clock = clock
            self.when = when
        def __call__(self, f, level=1):
            new_network_operation = NetworkOperation(f, self.clock, self.when)
            # Depending on whether we were called as @network_operation or
            # @network_operation(...) we need different levels, the level is
            # 2 in the first case and 1 in the second case (because in the
            # first case we go originalcaller->network_operation->do_network_operation
            # and in the second case we go originalcaller->do_network_operation
            # at the time when this method is called).
            new_network_operation.set_instance_id(level=level)
            new_network_operation.__name__ = f.__name__
            new_network_operation.__doc__ = f.__doc__
            new_network_operation.__dict__.update(f.__dict__)
            return new_network_operation
    if len(args) == 1 and callable(args[0]):
        # We're in case (1), the user has written:
        # @network_operation
        # def f():
        #    ...
        # and the single argument to the decorator is the function f
        return do_network_operation()(args[0], level=2)
    else:
        # We're in case (2), the user has written:
        # @network_operation(...)
        # def f():
        #    ...
        # and the arguments might be clocks or strings, and may have been
        # called with or without names, so we check both the variable length
        # argument list *args, and the keyword dictionary **kwds, falling
        # back on the default values if nothing is given.
        clk = None
        when = 'end'
        for arg in args:
            if isinstance(arg, Clock):
                clk = arg
            elif isinstance(arg, str):
                when = arg
        for key, val in kwds.iteritems():
            if key == 'clock': clk = val
            if key == 'when': when = val
        return do_network_operation(clock=clk, when=when)
    #raise TypeError, "Decorator must be used as @network_operation or @network_operation(clock)"


class MagicNetwork(Network):
    '''
    Creates a :class:`Network` object from any suitable objects
    
    **Initialised as:** ::
    
        MagicNetwork()
    
    The object returned can then be used just as a regular
    :class:`Network` object. It works by finding any object in
    the ''execution frame'' (i.e. in the same function, script
    or section of module code where the :class:`MagicNetwork` was
    created) derived from :class:`NeuronGroup`, :class:`Connection` or
    :class:`NetworkOperation`.
    
    **Sample usage:** ::
    
        G = NeuronGroup(...)
        C = Connection(...)
        @network_operation
        def f():
            ...
        net = MagicNetwork()
    
    Each of the objects ``G``, ``C`` and ``f`` are added to ``net``.
    
    **Advanced usage:** ::
    
        MagicNetwork([verbose=False[,level=1]])
    
    with arguments:
    
    ``verbose``
        Set to ``True`` to print out a list of objects that were
        added to the network, for debugging purposes.
    ``level``
        Where to find objects. ``level=1`` means look for objects
        where the :class:`MagicNetwork` object was created. The ``level``
        argument says how many steps back in the stack to look.
    '''
    def __init__(self, verbose=False, level=1):
        '''
        Set verbose=False to turn off comments.
        The level variable contains the location of the namespace.
        '''
        (groups, groupnames) = magic.find_instances(NeuronGroup)
        groups = [g for g in groups if g._owner is g]
        groupnames = [gn for g, gn in zip(groups, groupnames) if g._owner is g]
        (connections, connectionnames) = magic.find_instances(Connection)
        (operations, operationnames) = magic.find_instances(NetworkOperation)
        if verbose:
            print "[MagicNetwork] Groups:", groupnames
            print "[MagicNetwork] Connections:", connectionnames
            print "[MagicNetwork] Operations:", operationnames
        # Use set() to discard duplicates
        Network.__init__(self, list(set(groups)), list(set(connections)), list(set(operations)))


def run(duration, threads=1, report=None, report_period=10 * second):
    '''
    Run a network created from any suitable objects that can be found
    
    Arguments:
    
    ``duration``
        the length of time to run the network for.
    ``report``
        How to report progress, the default ``None`` doesn't report the
        progress. Some standard values for ``report``:
        
        ``text``, ``stdout``
            Prints progress to the standard output.
        ``stderr``
            Prints progress to the standard error output stderr.
        ``graphical``, ``tkinter``
            Uses the Tkinter module to show a graphical progress bar,
            this may interfere with any other GUI code you have.
            
        Alternatively, you can provide your own callback function by
        setting ``report`` to be a function ``report(elapsed, complete)``
        of two variables ``elapsed``, the amount of time elapsed in
        seconds, and ``complete`` the proportion of the run duration
        simulated (between 0 and 1). The ``report`` function is
        guaranteed to be called at the end of the run with
        ``complete=1.0`` so this can be used as a condition for
        reporting that the computation is finished.
    ``report_period``
        How often the progress is reported (by default, every 10s).
    
    Works by constructing a :class:`MagicNetwork` object from all the suitable
    objects that could be found (:class:`NeuronGroup`, :class:`Connection`, etc.) and
    then running that network. Not suitable for repeated runs or situations
    in which you need precise control.
    '''
    MagicNetwork(verbose=False, level=2).run(duration, threads=threads,
                                            report=report, report_period=report_period)


def reinit(states=True):
    '''
    Reinitialises any suitable objects that can be found
    
    **Usage:** ::
    
        reinit(states=True)
    
    Works by constructing a :class:`MagicNetwork` object from all the suitable
    objects that could be found (:class:`NeuronGroup`, :class:`Connection`, etc.) and
    then calling ``reinit()`` for each of them. Not suitable for repeated
    runs or situations in which you need precise control.
    
    If ``states=False`` then :class:`NeuronGroup` state variables will not be
    reinitialised.
    '''
    MagicNetwork(verbose=False, level=2).reinit(states=states)


def stop():
    '''
    Globally stops any running network, this is reset the next time a network is run
    '''
    global globally_stopped
    globally_stopped = True

def clear(erase=True, all=False):
    '''
    Clears all Brian objects.
    
    Specifically, it stops all existing Brian objects from being collected by
    :class:`MagicNetwork` (objects created after clearing will still be collected).
    If ``erase`` is ``True`` then it will also delete all data from these objects.
    This is useful in, for example, ``ipython`` which stores persistent references
    to objects in any given session, stopping the data and memory from being freed
    up.  If ``all=True`` then all Brian objects will be cleared. See also
    :func:`forget`.
    '''
    if all is False:
        net = MagicNetwork(level=2)
        objs = net.groups + net.connections + net.operations
    else:
        groups, _ = magic.find_instances(NeuronGroup, all=True)
        connections, _ = magic.find_instances(Connection, all=True)
        operations, _ = magic.find_instances(NetworkOperation, all=True)
        objs = groups+connections+operations
    for o in objs:
        o.set_instance_id(-1)
        if erase:
            for k, v in o.__dict__.iteritems():
                object.__setattr__(o, k, None)
    gc.collect()

def forget(*objs):
    '''
    Forgets the list of objects passed
    
    Forgetting means that :class:`MagicNetwork` will not pick up these objects,
    but all data is retained. You can pass objects or lists of objects. Forgotten
    objects can be recalled with :func:`recall`. See also :func:`clear`.
    '''
    for obj in objs:
        if isinstance(obj, (NeuronGroup, Connection, NetworkOperation)):
            obj._forgotten_instance_id = obj.get_instance_id()
            obj.set_instance_id(-1)
        elif isSequenceType(obj):
            for o in obj:
                forget(o)
        else:
            raise TypeError('Only the following types of objects can be forgotten: NeuronGroup, Connection or NetworkOperation')

def recall(*objs):
    '''
    Recalls previously forgotten objects
    
    See :func:`forget` and :func:`clear`.
    '''
    for obj in objs:
        if isinstance(obj, (NeuronGroup, Connection, NetworkOperation)):
            if hasattr(obj, '_forgotten_instance_id'):
                obj.set_instance_id(obj._forgotten_instance_id)
        elif isSequenceType(obj):
            for o in obj:
                recall(o)
        else:
            raise TypeError('Only the following types of objects can be recalled: NeuronGroup, Connection or NetworkOperation')



"""
A model of working memory from Wang (2002) implemented in Brian
"""
import time

time_start = time.time()
debugg = True      # set True for debugging (checks the units)

if not debugg:
  import brian_no_units

from brian import *


# === Define parameters
# size of the network
excit_n = 1600
inhib_n = 400

# simulation-related parameters
simtime = 1000.0 * ms     # total simulation time [ms]
dt = 0.02 * ms       # simulation step length [ms]

# === Parameters determining model dynamics =====
# pyramidal cells
Cm_e = 0.5 * nF     # [nF] total capacitance
gl_e = 25.0 * nS    # [ns] total leak conductance
El_e = -70.0 * mV   # [mV] leak reversal potential
Vth_e = -50.0 * mV   # [mV] threshold potential
Vr_e = -55.0 * mV   # [mV] resting potential
tr_e = 2.0 * ms   # [ms] refractory time

# interneuron cells
Cm_i = 0.2 * nF   # [nF] total capacitance
gl_i = 20.0 * nS   # [ns] total leak conductance
El_i = -70.0 * mV    # [mV] leak reversal potential
Vth_i = -50.0 * mV    # [mV] threshold potential
Vr_i = -55.0 * mV    # [mV] resting potential
tr_i = 1.0 * ms     # [ms] refractory time

#structure
N1 = 240
N2 = 2 * N1
f = 0.15
wp = 1.7
wm = 1.0 - f * (wp - 1.0) / (1.0 - f)

def structure(i, j):
    if(i == j): return 0
    if (i < N1):
        if (j < N1): return wp
        elif (j < N2): return wm
        else: return 1.0
    elif (i < N2):
        if (j < N1): return wm
        elif (j < N2): return wp
        else: return 1.0
    else:
        if (j < N2): return wm
        else: return 1.0

# external input
fext = 2400.0 * Hz   # [Hz] external input frequency (poisson train)
def stimulus(t):
  if (t < 1 * second): return fext
  elif (t < 1.5 * second): return 2440.0 * Hz
  else: return fext

# AMPA receptor (APMAR)
Vs_ampa = 0.0 * mV     # [mV] synaptic reversial potential
t_ampa = 2.0 * ms     # [ms] exponential decay time constant
g_ext_e = 2.1 * nS     # [nS] maximum conductance from external to pyramidal cells
g_ext_i = 1.62 * nS    # [nS] maximum conductance from external to interneuron cells
g_ampa_i = 0.04 * nS
g_ampa_e = 0.05 * nS

# GABA receptor (GABAR)
Vs_gaba = -70.0 * mV      # [mV] synaptic reversial potential
t_gaba = 5.0 * ms   # [ms] exponential decay time constant
g_gaba_e = 1.3 * nS
g_gaba_i = 1.0 * nS

# NMDA receptor (NMDAR)
Vs_nmda = 0.0 * mV     # [mV] synaptic reversial potential
ts_nmda = 100.0 * ms   # [ms] decay time of NMDA currents
tx_nmda = 2.0 * ms   # [ms] controls the rise time of NMDAR channels
alfa_nmda = 0.5 * kHz      # [kHz] controls the saturation properties of NMDAR channels
g_nmda_e = 0.165 * nS
g_nmda_i = 0.13 * nS


#################################

eqs_e = '''
dv/dt = (-gl_e*(v-El_e)-g_ext_e*(v-Vs_ampa)*s_ext-g_gaba_e*(v-Vs_gaba)*s_gaba-g_ampa_e*(v-Vs_ampa)*s_ampa-s_tot*g_nmda_e*(v-Vs_nmda)/(1+exp(-0.062*v/(1*mV))/3.57))/Cm_e: volt
ds_ext/dt = -s_ext/(t_ampa) : 1
ds_gaba/dt = -s_gaba/(t_gaba) : 1
ds_ampa/dt = -s_ampa/(t_ampa) : 1
ds_nmda/dt = -s_nmda/(ts_nmda)+alfa_nmda*x*(1-s_nmda) : 1
dx/dt = -x/(tx_nmda) : 1
s_tot : 1
'''
eqs_i = '''
dv/dt = (-gl_i*(v-El_i)-g_ext_i*(v-Vs_ampa)*s_ext-g_gaba_i*(v-Vs_gaba)*s_gaba-g_ampa_i*(v-Vs_ampa)*s_ampa-s_tot*g_nmda_i*(v-Vs_nmda)/(1+exp(-0.062*v/(1*mV))/3.57))/Cm_i: volt
ds_ext/dt = -s_ext/(t_ampa) : 1
ds_gaba/dt = -s_gaba/(t_gaba) : 1
ds_ampa/dt = -s_ampa/(t_ampa) : 1
s_tot : 1
'''

simulation_clock = Clock(dt=dt)

# Setting up the populations
print "Setting up the populations ..."


Pe = NeuronGroup(excit_n, eqs_e, threshold=Vth_e, reset=Vr_e, refractory=tr_e, order=2)
Pe.v = El_e
Pe.s_ext = 0
Pe.s_ampa = 0
Pe.s_gaba = 0
Pe.s_nmda = 0
Pe.x = 0

Pi = NeuronGroup(inhib_n, eqs_i, threshold=Vth_i, reset=Vr_i, refractory=tr_i, order=2)
Pi.v = El_i
Pi.s_ext = 0
Pi.s_ampa = 0
Pi.s_gaba = 0

#3 excitatory subgroups: 1 & 2 are selective to motion, ns not
Pe_1 = Pe.subgroup(N1)
Pe_2 = Pe.subgroup(N1)
Pe_ns = Pe.subgroup(excit_n - 2 * N1)

PG1 = PoissonGroup(N1, lambda t:stimulus(t))
PG2 = PoissonGroup(N1, lambda t:stimulus(t))
PGns = PoissonGroup(excit_n - 2 * N1, fext)
PGi = PoissonGroup(inhib_n, fext)

# Creating connections
print "Creating static conections ..."

selfnmda = IdentityConnection(Pe, Pe, 'x', weight=1.0, delay=0.5 * ms)

Cp1 = IdentityConnection(PG1, Pe_1, 's_ext', weight=1.0)
Cp2 = IdentityConnection(PG2, Pe_2, 's_ext', weight=1.0)
Cpns = IdentityConnection(PGns, Pe_ns, 's_ext', weight=1.0)
Cpi = IdentityConnection(PGi, Pi, 's_ext', weight=1.0)

Cie = Connection(Pi, Pe, 's_gaba', weight=1.0, delay=0.5 * ms)
Cii = Connection(Pi, Pi, 's_gaba', weight=1.0, delay=0.5 * ms)
Cei = Connection(Pe, Pi, 's_ampa', weight=1.0, delay=0.5 * ms)
Cee = Connection(Pe, Pe, 's_ampa', weight=structure, delay=0.5 * ms)

for i in range(len(Pi)):
  Cii[i, i] = 0

Wee = zeros((excit_n, excit_n))
for i in range(len(Pe)):
  for j in range(len(Pe)):
   Wee[i, j] = structure(i, j)

Wei = ones((excit_n, inhib_n))
Wei = transpose(Wei)

@network_operation(simulation_clock, when='start')
def f(simulation_clock):
    s_NMDA_full = transpose(array([Pe.s_nmda]))   # get the s_NMDA values of the excitatory neurons
    Pe.s_tot = transpose(dot(Wee, s_NMDA_full))  # Calculate the new s_tot values based on the weights for  the excitatory neurons
    Pi.s_tot = transpose(dot(Wei, s_NMDA_full))  # Calculate the new s_tot values based on the weights for the inhibitory neurons

# initiating monitors
M = SpikeMonitor(Pe)
#N = SpikeMonitor(Pi)

print "Running ..."
run(simtime, report='text')

print "Simulation finished."
raster_plot(M)
#raster_plot(N)
show()



from brian import *
from customrefractoriness import *
from scipy.optimize import fmin
from brian.tools.autodiff import *

#----Fixed parameters
ENa = 60 * mV
El = -70 * mV
tau = 5 * ms
tauh = 5 * ms

#----Input parameters
mu = 0 * mV
sigma = 2 * 15 * mV
tauc = 10 * ms
k = 1 / 3 * sigma / ms                                                      #empirical threshold criterion

#----Na channels parameters
#--load the data
#Va,Ka,Vi,Ki,R=load('Nav1_6.txt').T#I don't have this file
Va, Ka, Vi, Ki, R = (zeros(10), zeros(10), zeros(10), zeros(10), zeros(10))
N = len(Va)

eqs = Equations("""
# Sodium channel model
Pa=1/(1+exp(-(v-va)/ka)) : 1
Pi=1/(1+exp(-(v-vi)/ki)) : 1

# Membrane equation
dv/dt=(-r*Pa*h*(v-ENa)-(v-El)+I)/tau : volt
dh/dt=((1-Pi)-h)/tauh : 1

# Input
dI/dt=(mu-I)/tauc+sigma*(.5*tauc)**-.5*xi : volt

# Sodium channel model Parameters
va : mV
ka : mV
vi : mV
ki : mV
r : 1
""")
eqs.prepare()

#--Function for computing the EIF and threshold dynamics parameters
def threshold_parameters(f, h_inf, rest=El):
   '''
   Calculates the threshold and the slope factor from the
   equation tau*dv/dt=f(v)
   '''
   vt = fmin(f, rest, disp=False)[0] # Threshold (simplex algorithm)
   vt = vt * volt
   deltat = 1 / (tau * differentiate(f, vt, order=2)) # Slope factor
   a = -differentiate(h_inf, vt, order=1) * deltat / h_inf(vt)
   b = -deltat * math.log(h_inf(vt)) + (1 - a) * vt
   return (vt, deltat, a, b)

#--Predictions for each Na channel model
vt = [[] for i in range(len(Va))]
deltat = [[] for i in range(len(Va))]
a = [[] for i in range(len(Va))]
b = [[] for i in range(len(Va))]
pente = [[] for i in range(len(Va))]
ordonnee = [[] for i in range(len(Va))]
figure()
for j in range(len(Va)):
    print "Sodium channel model", j + 1
    print "----------------------"
    va = Va[j] * mV
    ka = Ka[j] * mV
    vi = Vi[j] * mV
    ki = Ki[j] * mV
    r = R[j]
    print "Va =", va, "mV ; ka =", ka, "mV ;", "Vi =", vi, "mV ; ki =", ki, "mV"
#--Boltzmann functions
    Pa = lambda v: eqs.apply('Pa', {'v':v, 'va':va, 'ka':ka})                            # activation
    h_inf = lambda v: 1 - eqs.apply('Pi', {'v':v, 'vi':vi, 'ki':ki})                           # inactivation
#--The EIF model   
    # F=lambda v: eqs.apply('v',{'Pa':Pa(v),'r':r,'h':1,'I':mu,'v':v})  # TODO: Romain, this doesn't work and probably should, see note in Equations.apply()
    F = lambda v: eqs.apply('v', {'r':r, 'h':1, 'I':mu, 'v':v, 'va':va, 'ka':ka})
    v = linspace(-100 * mV, 0 * mV, 200)
    plot(v / mV, Pa(v))
    plot(v / mV, F(v) * ms / mV)
show()



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
'''
Tabulation of numerical functions.
'''

__all__ = ['Tabulate', 'TabulateInterp']

from brian.units import get_unit, Quantity, is_dimensionless
from brian.unitsafefunctions import array, arange, zeros
from numpy import NaN


class Tabulate(object):
    '''
    An object to tabulate a numerical function.
    
    Sample use::
    
      g=Tabulate(f,0.,1.,1000)
      y=g(.5)
      v=g([.1,.3])
      v=g(array([.1,.3]))
      
    Arguments of g must lie in [xmin,xmax).
    An IndexError is raised is arguments are above xmax, but
    not always when they are below xmin (it can give weird results).
    '''
    def __init__(self, f, xmin, xmax, n):
        self.xmin = xmin
        self.xmax = xmax
        self.dx = (xmax - xmin) / float(n)
        self.invdx = 1 / self.dx
        self.unit = get_unit(f(xmin))
        # Tabulation at midpoints
        x = xmin + (.5 + arange(n)) * self.dx
        try:
            self.f = f(x)
        except:
            # If it fails we try passing the values one by one
            self.f = zeros(n) * f(xmin) # for the unit
            for i in xrange(n):
                self.f[i] = f(x[i])

    def __call__(self, x):
        try: # possible problem if x is an array and an array is wanted
            return self.f[array((array(x) - self.xmin) * self.invdx, dtype=int)]
        except IndexError: # out of bounds
            return NaN * self.unit

    def __repr__(self):
        return 'Tabulated function with ' + str(len(self.f)) + ' points'


class TabulateInterp(object):
    '''
    An object to tabulate a numerical function with linear interpolation.
    
    Sample use::
    
      g=TabulateInterp(f,0.,1.,1000)
      y=g(.5)
      v=g([.1,.3])
      v=g(array([.1,.3]))
      
    Arguments of g must lie in [xmin,xmax).
    An IndexError is raised is arguments are above xmax, but
    not always when they are below xmin (it can give weird results).
    '''
    def __init__(self, f, xmin, xmax, n):
        self.xmin = xmin
        self.xmax = xmax
        self.dx = (xmax - xmin) / float(n - 1)
        self.invdx = 1 / self.dx
        # Not at midpoints here
        x = xmin + arange(n) * self.dx
        self.unit = get_unit(f(xmin))
        try:
            self.f = f(x)
        except:
            # If it fails we try passing the values one by one
            self.f = zeros(n) * f(xmin) # for the unit
            for i in xrange(n):
                self.f[i] = f(x[i])
        self.f = array(self.f)
        self.df = (self.f[range(1, n)] - self.f[range(n - 1)]) * float(self.invdx)

    def __call__(self, x): # the units of x is not checked
        y = array(x) - self.xmin
        ind = array(y * self.invdx, dtype=int)
        try:
            if is_dimensionless(x): # could be a problem if it is a Quantity with units=1
                return self.f[ind] + self.df[ind] * (y - array(ind) * self.dx)
            else:
                return array(self.f[ind] + self.df[ind] * (y - array(ind) * self.dx)) * self.unit
        except IndexError: # out of bounds
            return NaN * self.unit

    def __repr__(self):
        return 'Tabulated function with ' + str(len(self.f)) + ' points (interpolated)'



'''
Simple pickling example with a small network.
This script looks for a previous data file
and if it finds one reads in saved network parameters
and connection weights.
Network parameters and connection weights are saved
at the end of a run.
'''
from brian import *
import pickle

# Look for an import file - if not there start
# a new empty one

try:
    imported_data = pickle.load(open('./saved.pickle', 'rb'))
    defaultclock.t = imported_data['t']
    print 'Starting from saved progress at time', defaultclock.t

except IOError:
    print 'No data to import, starting new run'
    imported_data = {}

# function which saves the network and connection
# data using pickle.

def save_progress():
    s = str(int((defaultclock.t + .5 * ms) / second))
    imported_data['P._S'] = P._S
    imported_data['t'] = defaultclock.t
    exc_inds = [Ce.W[i, :].ind for i in range(Ne)]
    exc_weights = [asarray(Ce.W[i, :]) for i in range(Ne)]
    inh_inds = [Ci.W[i, :].ind for i in range(Ni)]
    inh_weights = [asarray(Ci.W[i, :]) for i in range(Ni)]
    imported_data['exc_inds'] = exc_inds
    imported_data['exc_weights'] = exc_weights
    imported_data['inh_inds'] = inh_inds
    imported_data['inh_weights'] = inh_weights
    pickle.dump(imported_data, open('./saved.pickle', 'wb'), -1)
    print 'Saved progress up to time', s, 'second'

# test network function using a different clock
# from default
@network_operation(clock=EventClock(t=defaultclock.t, dt=1000 * msecond))
def test_func():
    print "Executing test_func at ", defaultclock.t


# Network parameters
Ne = 200
Ni = 50

# Model equations
eqs = '''
dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
dge/dt = -ge/(5*ms) : volt
dgi/dt = -gi/(10*ms) : volt
'''

# Create Neuron group
P = NeuronGroup(Ne + Ni, model=eqs, threshold= -50 * mV, reset= -60 * mV)
P.v = -60 * mV
Pe = P.subgroup(Ne)
Pi = P.subgroup(Ni)

# If initialising from previous data then read in Neuron
# group data
if imported_data:
    P._S[:] = imported_data['P._S']

# Create connections

Ce = Connection(Pe, P, 'ge')
Ci = Connection(Pi, P, 'gi')

# If initialising from previous data then read in Connection
# data, otherwise connect randomly

if imported_data:
    exc_inds = imported_data['exc_inds']
    exc_weights = imported_data['exc_weights']
    inh_inds = imported_data['inh_inds']
    inh_weights = imported_data['inh_weights']
    for i in xrange(Ne):
        Ce[i, exc_inds[i]] = exc_weights[i]
    for i in xrange(Ni):
        Ci[i, inh_inds[i]] = inh_weights[i]
else:
    Ce.connect_random(Pe, P, 0.02, weight=1.62 * mV)
    Ci.connect_random(Pi, P, 0.02, weight= -9 * mV)


# Do the run
M = SpikeMonitor(P)

run(4000 * ms)

# Save progress after run
save_progress()

# show what we've computed so far...
raster_plot(M)
show()



import brian_no_units
from brian import *
import time

#set_global_preferences(useweave=False)
print 'Compilation on', get_global_preference('useweave')

duration = 1 * second
Nsyn = 80.
we = 1.62 * mV
N = 512 * 16
structure = 'dense'

eqs = '''
dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
dge/dt = -ge/(5*ms) : volt
dgi/dt = -gi/(10*ms) : volt
'''

Ne = int(N * 0.8)
Ni = N - Ne

P = NeuronGroup(N, model=eqs,
              threshold= -50 * mV, reset= -60 * mV, method='Euler', compile=True, freeze=True)#, refractory=5*ms)
P.v = -60 * mV + 10 * mV * rand(len(P))
Pe = P.subgroup(Ne)
Pi = P.subgroup(Ni)

Ce = Connection(Pe, P, 'ge', structure=structure)
Ci = Connection(Pi, P, 'gi', structure=structure)
cp = Nsyn / N
Ce.connect_random(Pe, P, cp, weight=we)
Ci.connect_random(Pi, P, cp, weight= -9 * mV)

M = PopulationSpikeCounter(P)

P.v = -60 * mV + 10 * mV * rand(len(P))

#Ce.W.freeze()
#Ci.W.freeze()

run(0 * ms)

def f():
    tstart = time.time()
    run(duration)
    tend = time.time()

    print 'N:', N
    print 'CPU time:', tend - tstart
    print 'Num spikes', M.nspikes

f()
#import cProfile as profile
#import pstats
#profile.run('f()','cuba_once.prof')
#stats = pstats.Stats('cuba_once.prof')
##stats.strip_dirs()
#stats.sort_stats('cumulative', 'calls')
#stats.print_stats(50)



'''
Comparison of two algorithms for the new spike/refractoriness mechanism. These
algorithms are capable of arbitrarily long refractory periods which don't need
to be specified in advance, can change from moment to moment (although once a
spike is fired the refractory period for that firing is fixed), and can signal
differently for each condition (1) spike fired, (2) in refractory period,
(3) leaving refractory period. These could be used by the refractoriness
mechanism in different ways.

Results:

N = 100, duration = 10000, thr=0.5, refracperiod = arange(N)
algo 1: .7
algo 2: .96

N = 100, duration = 10000, thr=0.5, refracperiod = zeros(N)
algo 1: .72
algo 2: .97

N = 100, duration = 10000, thr=0.5, refracperiod = 10*ones(N)
algo 1: .75
algo 2: .99

N = 100, duration = 10000, thr=0.5, refracperiod = 500*ones(N)
algo 1: .7
algo 2: .95

So it looks like distribution of refrac periods doesn't count much.

N = 100, duration = 10000, thr=0.1, refracperiod = arange(N)
algo 1: .69
algo 2: .91

N = 100, duration = 10000, thr=0.01, refracperiod = arange(N)
algo 1: .67
algo 2: .875

N = 100, duration = 10000, thr=0.9, refracperiod = arange(N)
algo 1: .72
algo 2: .97

So it looks like firing rate doesn't matter much.

N = 1000, duration = 1000, thr=0.5, refracperiod = arange(N)
algo 1: .19
algo 2: .24

N = 10000, duration = 10000, thr=0.5, refracperiod = arange(N)
algo 1: 14.3
algo 2: 16.5

So it looks like number of neurons doesn't make much difference.

Conclusion: algo 1 > algo 2

For large N, it appears to take about as long to do refractoriness as it does
to do thresholding. I think this seems relatively inexpensive.
'''

from brian import *
import time

N = 10000
duration = 10000
thr = 0.5
dorecord = False
algo = 1

refracperiod = zeros(N, dtype=int)#arange(N)

staterec = []
active = ones(N, dtype=bool)

start = time.time()
for t in xrange(duration):
    threshold = (rand(N)<thr).nonzero()[0]
    spikes = threshold[active[threshold]]
end = time.time()

t_thresh = end-start
print 'Time spent doing thresholding:', t_thresh

if algo==1:

    refend = zeros(N, dtype=int)
    ref = zeros(N, dtype=int)
    numref = 0

    start = time.time()
    for t in xrange(duration):
        if dorecord: state = zeros(N)
        threshold = (rand(N)<thr).nonzero()[0]
        spikes = threshold[active[threshold]]
        nspikes = len(spikes)
        active[spikes] = False
        ref[numref:numref+nspikes] = spikes
        refend[numref:numref+nspikes] = t+refracperiod[spikes]
        if dorecord: state[ref[:numref]] = 2
        numref += nspikes
        # sorted(ref[:numref]) # at moderate cost (an additional 30%) can return sorted array, which might be better for cache?
        # on the other hand, cache may not be so important unless a large number of neurons are refractory. Probably better to leave
        # this out overall.
        if dorecord: state[spikes] = 1
        I = (t>=refend[:numref])
        notI = -I
        leaving = ref[I.nonzero()[0]]
        staying = notI.nonzero()[0]
        active[leaving] = True
        if dorecord: state[leaving] = 3
        nstaying = len(staying)
        ref[:nstaying] = ref[staying]
        refend[:nstaying] = refend[staying]
        numref = nstaying
        if dorecord: state += active*0.3
        if dorecord: staterec.append(state)
    end = time.time()

elif algo==2:

    refend = zeros(N, dtype=int)
    ref = zeros(0, dtype=int)
    numref = 0

    start = time.time()
    for t in xrange(duration):
        if dorecord: state = zeros(N)
        threshold = (rand(N)<thr).nonzero()[0]
        spikes = threshold[active[threshold]]
        if dorecord: state[ref] = 2
        ref = union1d(ref, spikes) # this involves doing a sort operation, although it could be accelerated with weave easily enough
        active[spikes] = False
        refend[spikes] = t+refracperiod[spikes]
        if dorecord: state[spikes] = 1
        I = (t>=refend[ref])
        notI = -I
        leaving = ref[I]
        active[leaving] = True
        if dorecord: state[leaving] = 3
        ref = ref[notI]
        if dorecord: state += active*0.3
        if dorecord: staterec.append(state)
    end = time.time()
    
print 'Time spent total:', end-start
print 'Time spent on algorithm:', end-start-t_thresh

if dorecord:
    imshow(array(staterec).T, interpolation='nearest', origin='bottom left', aspect='auto')
    colorbar()
    show()
